"""
File containing a class for checking a device (or/and a surrogate model) against previously obtained reference measurements.
"""
import os
import time
import torch
import numpy as np
import matplotlib.pyplot as plt
from brainspy.utils.io import load_configs
from bspysmg.data.sampling import Sampler
from brainspy.utils.io import create_directory_timestamp
from brainspy.utils.pytorch import TorchUtils
from typing import Tuple


class ConsistencyChecker(Sampler):
    def __init__(self,
                 main_dir: str,
                 repetitions: int = 1,
                 sampler_configs_name: str = 'sampler_configs.json',
                 reference_batch_name: str = 'reference_batch.npz',
                 charging_signal_name: str = 'charging_signal.npz',
                 model: torch.nn.Module = None,
                 show_plots: bool = True) -> None:
        """
        Initializes dataset and directory to save results for consistency checking
        experiment of a model. This function uses sampler config files and already
        existing device input-output dataset and original measurements.

        Parameters
        ----------
        main_dir : str
            Path to main directory which contains the configuration files.
        repetitions : int [Optional]
            Number of times the experiments should be repeated.
        sampler_configs_name : str [Optional]
            Name of the file which contains sampling configuration and is used to initialize the
            parent class Sampler. It has the following keys:
            * save_directory: str
                Directory where the all the sampling data will be stored.
            * data_name: str
                Inside the path specified on the variable save_directory, a folder will be created,
                with the format: <data_name>+<current_timestamp>. This variable specified the
                prefix of that folder before the timestamp.
            * driver: dict
                Dictionary containing the driver configurations. For more information check the
                documentation about this configuration file, check the documentation of
                brainspy.processors.hardware.drivers.ni.setup.NationalInstrumentsSetup
            * input_data : dict
                Dictionary containing the information necessary to create the input sampling data.
                - input_distribution: str
                    It determines the wave shape of the input. Two main options availeble 'sawtooth'
                    and 'sine'. The first option will create saw-like signals, and the second
                    sine-wave signals. Sawtooth signals have more coverage on the edges of the
                    input range.
                - activation_electrode_no: int
                    Number of activation electrodes in the device that wants to be sampled.
                - readout_electrode_no : int
                    Number of readout electrodes in the device that wants to be sampled.
                - input_frequency: list
                    Base frequencies of the input waves that will be created. In order to optimise
                    coverage, irrational numbers are recommended. The list should have the same
                    length as the activation electrode number. E.g., for 7 activation electrodes:
                    input_frequency = [2, 3, 5, 7, 13, 17, 19]
                - phase : float
                    Horizontal shift of the input signals. It is recommended to have random numbers
                    which are different for the training, validation and test datasets. These
                    numbers will be square rooted and multiplied by a given factor.
                - factor : float
                    Given factor by which the input frequencies will be multiplied after square
                    rooting them.
                - amplitude : Optional[list[float]]
                    Amplitude of the generated input wave signal. It is calculated according to the
                    minimum and maximum ranges of each electrode. Where the amplitude value should
                    correspond with (max_range_value - min_range_value) / 2. If no amplitude is
                    given it will be automatically calculated from the driver configurations for
                    activation electrode ranges. If it wants to be manually set, the offset
                    variable should also be included in the dictionary.
                - offset: Optional[list[float]]
                    Vertical offset of the generated input wave signal. It is calculated according
                    to the minimum and maximum ranges of each electrode. Where the offset value
                    should correspond with (max_range_value + min_range_value) / 2. If no offset
                    is given it will be automatically calculated from the driver configurations for
                    activation electrode ranges. If it wants to be manually set, the offset
                    variable should also be included in the dictionary.
                - ramp_time: float
                    Time that will be taken before sending each batch to go from zero to the first
                    point of the batch and to zero from the last point of the batch.
                - batch_time:
                    Time that the sampling of each batch will take.
                - number_batches: int
                    Number of batches that will be sampled. A default value of 3880 is reccommended.
        reference_batch_name : str [Optional]
            Name of the file which contains the reference dataset. This is the original device
            measurements. It is a npz file with columns 'inputs' and 'outputs'.
        charging_signal_name : str [Optional]
            Name of the file which contains device inputs and outputs. This is the device behaviour
            at present moment. It is a npz file with columns 'inputs' and 'outputs'.
        model : custom model of type torch.nn.Module [Optional]
            Model whose consistency is to be checked. This is a trained neural network model over
            DNPU measurements and sampled input data.
        """
        super(ConsistencyChecker, self).__init__(
            load_configs(os.path.join(main_dir, sampler_configs_name)))
        _, self.batch_size, _ = self.init_configs()
        path_to_file = os.path.join(main_dir, reference_batch_name)
        with np.load(path_to_file) as data:
            self.reference_outputs = data['outputs']
            self.reference_inputs = data['inputs'].T
        path_to_file = os.path.join(main_dir, charging_signal_name)
        with np.load(path_to_file) as data:
            self.chargingup_outputs = data['outputs']
            self.chargingup_inputs = data['inputs'].T
        self.nr_samples = len(self.reference_outputs)
        assert self.nr_samples % self.batch_size == 0, f"Batch length {self.batch_size} is not a multiple of the reference signal length {self.nr_samples}; possible data missmatch!"
        self.results_dir = create_directory_timestamp(main_dir,
                                                      'consistency_check')
        self.results_filename = os.path.join(self.results_dir,
                                             'consistency_results.npz')
        self.repetitions = repetitions
        self.model = model

    def get_data(self, charge_device=True) -> Tuple[np.array]:
        """
        The main function that implements consistency checking routine. It uses
        the reference data and device's outputs to check if the outputs of device
        are consistent with device over several runs. Optionally it can also check
        consistency of a trained neural network over device measurements.

        Returns
        -------
        tuple
            tuple with following data:
                - results: np.array
                    outputs generated by the device.
                - deviations: np.array
                    RMSE between device outputs and refernce data.
                - correlation: np.array
                    Value of correlation coefficient between device outputs and
                    refence data.
                - deviation_chargeup: np.array
                    RMSE deviation between device output and original device output.
                
                if model is not None:
                    - model_results: np.array
                        outputs generated by the model.
                    - model_deviations: np.array
                        RMSE between model outputs and refernce data.
                    - model_correlation: np.array
                        Value of correlation coefficient between model outputs and
                        refence data.
                    - model_deviation_chargeup: np.array
                        RMSE deviation between model output and original device output.
        """
        results = np.zeros((self.repetitions, ) + self.reference_outputs.shape)
        deviations = np.zeros(self.repetitions)
        correlation = np.zeros(self.repetitions)
        deviation_chargeup = []
        if self.model is not None:
            model_results = results.copy()
            model_deviations = deviations.copy()
            model_correlation = correlation.copy()
            model_deviation_chargeup = []
            if charge_device:
                self.charge_device(deviation_chargeup,
                                   model_deviation_chargeup)
        else:
            if charge_device:
                self.charge_device(deviation_chargeup)
        # Initialize sampling loop
        for trial in range(self.repetitions):
            start_trial = time.time()
            for batch, batch_indices in enumerate(
                    self.get_batch_indices(self.nr_samples, self.batch_size)):

                # Generate inputs (without ramping)
                inputs = self.reference_inputs[:, batch_indices]
                # Get outputs (without ramping; raming is done in the get_batch method)
                device_outputs = self.sample_batch(inputs)
                results[trial, batch_indices] = device_outputs
                # if (batch % 1) == 0:
                #     self.plot_waves(inputs.T, outputs, batch)
                if self.model is not None:
                    model_outputs = self.sample_model_batch(inputs)
                    model_results[trial, batch_indices] = model_outputs
            end_trial = time.time()
            deviations[trial] = np.sqrt(
                np.mean((results[trial] - self.reference_outputs)**2))
            correlation[trial] = np.corrcoef(results[trial].T,
                                             self.reference_outputs.T)[0, 1]
            print(
                f'* Consistency check {trial+1}/{self.repetitions} took {end_trial - start_trial :.2f} sec. with {batch+1} batches'
            )
            print(
                f"\tCorr: {correlation[trial]:.2f} ; RMSE Deviation: {deviations[trial]:.2f}"
            )
            if self.model is not None:
                model_deviations[trial] = np.sqrt(
                    np.mean((results[trial] - model_results[trial])**2))
                model_correlation[trial] = np.corrcoef(
                    results[trial].T, model_results[trial].T)[0, 1]
                print(
                    f"\tCorr (Model): {model_correlation[trial]:.2f} ; RMSE Deviation (Model): {model_deviations[trial]:.2f}"
                )
        self.driver.close_tasks()
        #self.close_processor()

        np.savez(self.results_filename,
                 results=results,
                 deviations=deviations,
                 correlation=correlation)
        print(f'Data saved in { self.results_filename}')
        if self.model is None:
            return results, deviations, correlation, deviation_chargeup
        else:
            return results, deviations, correlation, deviation_chargeup, model_results, model_deviations, model_correlation, model_deviation_chargeup

    def charge_device(self, deviation_chargeup, model_deviation_chargeup=None, show_plots=True):
        for batch, batch_indices in enumerate(
                self.get_batch_indices(len(self.chargingup_outputs),
                                       self.batch_size)):
            # Generate inputs (without ramping)
            inputs = self.chargingup_inputs[:, batch_indices]

            # Get outputs (without ramping)
            outputs = self.sample_batch(inputs)
            charging_signal_deviations = np.sqrt(
                np.mean((outputs - self.chargingup_outputs[batch_indices])**2))
            deviation_chargeup.append(charging_signal_deviations)
            print(
                f'\n* Charging up: Batch {batch+1}/{int(len(self.chargingup_outputs)/self.batch_size)}\n\tRMSE deviation of measured device output against original device output: {charging_signal_deviations:.2f} (nA)'
            )
            if self.model is not None:
                model_outputs = self.sample_model_batch(inputs)
                model_charging_signal_deviations = np.sqrt(
                    np.mean((outputs - model_outputs)**2))
                original_deviation = np.sqrt(
                    np.mean((self.chargingup_outputs[batch_indices] -
                             model_outputs)**2))
                model_deviation_chargeup.append(
                    model_charging_signal_deviations)
                print(
                    f'\tRMSE deviation of measured device output against model output: {model_charging_signal_deviations:.2f} (nA)'
                )
                print(
                    f'\tRMSE deviation of original device output against model output: {original_deviation:.2f} (nA)'
                )
            if batch == 0:
                plt.plot(self.chargingup_outputs[batch_indices],
                         label='Original output',
                         alpha=0.5)
                plt.plot(outputs, label='Measured output', alpha=0.5)
                if self.model is not None:
                    plt.plot(model_outputs, label='Model output', alpha=0.5)
                plt.title(f'Charging signal (Batch 0)')
                plt.legend()
                plt.savefig(os.path.join(self.results_dir, 'first_batch'))
        print('Finished charging up device. \n')

    # def get_batch(self, input_batch):
    #      super(ConsistencyChecker, self).get_batch(input_batch)
    #     # # Ramp input batch (0.5 sec up and down)
    #     # batch_ramped = self.ramp_input_batch(input_batch)
    #     # # Readout output signal
    #     # outputs_ramped = self.driver.forward_numpy(batch_ramped.T)
    #     return outputs_device[self.filter_ramp]

    def sample_model_batch(self, input_batch: torch.Tensor) -> torch.Tensor:
        """
        Ramp the input batch (0.5 sec up and down) and format it to Tensor. Ramping is
        required to avoid abrupt changes in the voltage. The input is masked from 0v to
        first value and then final value back to 0v.

        Returns
        -------
        torch.Tensor
            Input tensor ramped and allocated to tensor.
        """
        #outputs_device = super(ConsistencyChecker,self).get_batch(input_batch)
        # outputs_device = super(ConsistencyChecker,self).sample_batch(input_batch)
        # Ramp input batch (0.5 sec up and down) and format it to pytorch
        #batch_ramped = TorchUtils.format(self.ramp_input(input_batch).T)
        self.model.eval()
        with torch.no_grad():
            outputs = TorchUtils.to_numpy(
                self.model(TorchUtils.format(input_batch).T))  #.squeez0e(0)
        # if len(outputs_ramped.shape) > 1:
        #     return outputs_ramped[self.filter_ramp[:, np.newaxis]][:,
        #                                                            np.newaxis]
        return outputs


def consistency_check(main_dir: str,
                      repetitions: int = 1,
                      sampler_configs_name: str = 'sampler_configs.json',
                      reference_batch_name: str = 'reference_batch.npz',
                      charging_signal_name: str = 'charging_signal.npz',
                      charge_device: bool = True,
                      model: torch.nn.Module = None,
                      show_plots = True) -> None:
    """
    This is the driver function used for consistency checking. Consistency checking involves
    checking how DNPU device is behaving at present moment against how it was before
    measurement. This check can also be performed against trained neural network model over
    DNPU measurements. This function initializes a ConsistencyChecker object and performs the 
    consistency check using the get_data function. It also plots and saves various graphs of
    calculated metrics.

    Parameters
    ----------
    main_dir : str
        Path to main directory which contains the configuration files.
    repetitions : int [Optional]
        Number of times the experiments should be repeated.
    sampler_configs_name : str [Optional]
        Name of the file which contains sampling configuration with keys mentioned in
        constructor function of ConsistencyChecker class.
    reference_batch_name : str [Optional]
        Name of the file which contains the reference dataset. This is the original device
        measurements. It is a npz file with columns 'inputs' and 'outputs'.
    charging_signal_name : str [Optional]
        Name of the file which contains device inputs and outputs. This is the device behaviour
        at present moment. It is a npz file with columns 'inputs' and 'outputs'.
    charge_device: boolean [Optional]
        Whether the consistency check should charge up the device with the charging signal or not.
    model : custom model of type torch.nn.Module [Optional]
        Model whose consistency is to be checked. This is a trained neural network model over
        DNPU measurements and sampled input data.
    """
    sampler = ConsistencyChecker(main_dir,
                                 repetitions=repetitions,
                                 sampler_configs_name=sampler_configs_name,
                                 reference_batch_name=reference_batch_name,
                                 charging_signal_name=charging_signal_name,
                                 model=model)
    if model is None:
        outputs, deviations, correlation, deviation_chargeup = sampler.get_data(
            charge_device)
    else:
        outputs, deviations, correlation, deviation_chargeup, model_outputs, model_deviations, model_correlation, model_deviation_chargeup = sampler.get_data(
            charge_device)
    mean_output = np.mean(outputs, axis=0)
    std_output = np.std(outputs, axis=0)
    plt.figure()
    plt.plot(sampler.reference_outputs,
             'r',
             label='Reference signal',
             alpha=0.5)
    plt.plot(mean_output, 'k', label='Device output (mean)', alpha=0.5)
    plt.plot(mean_output + std_output, ':k', alpha=0.5)
    plt.plot(mean_output - std_output,
             ':k',
             label='Device output (std)',
             alpha=0.5)
    if model is not None:
        model_mean_output = np.mean(model_outputs, axis=0)
        model_std_output = np.std(model_outputs, axis=0)
        plt.plot(model_mean_output,
                 'c',
                 label='Model output (mean)',
                 alpha=0.5)
        # It could be removed with if model.noise is None
        plt.plot(model_mean_output + model_std_output, ':c', alpha=0.5)
        plt.plot(model_mean_output - model_std_output,
                 ":c",
                 label='Model output (std)')
    plt.title(f'Reference signal over {sampler.repetitions} trials')
    plt.legend()
    plt.savefig(os.path.join(sampler.results_dir, 'consistency_check'))

    plt.figure()
    plt.hist(np.sqrt((mean_output - sampler.reference_outputs)**2),
             bins=100,
             alpha=0.5,
             label='Device')
    if model is not None:
        plt.hist(np.sqrt((model_mean_output - sampler.reference_outputs)**2),
                 bins=100,
                 alpha=0.5,
                 label="Model")
    plt.title(
        "Reference Signal\nHistogram of RMSE deviations (nA) from mean over " +
        str(repetitions) + " trials")
    plt.legend()
    # plt.figure()
    # plt.plot(mean_output - sampler.reference_outputs,
    #          "b",
    #          label="Error over mean (Device)", alpha=0.5)
    # plt.plot(mean_output - sampler.reference_outputs + std_output,
    #          ":b",
    #          label="Error over std (Device) ", alpha=0.5)
    # plt.plot(mean_output - sampler.reference_outputs - std_output, ":b"
    # , alpha=0.5)
    # if model is not None:
    #     plt.plot(model_mean_output - sampler.reference_outputs, "c",
    #     label="Error over mean (Model)", alpha=0.5)
    #     plt.plot(model_mean_output - sampler.reference_outputs + model_std_output,
    #     ":c",
    #     label="Error over std (Model) ", alpha=0.5)
    #     plt.plot(model_mean_output - sampler.reference_outputs + model_std_output,
    #     ":c", alpha=0.5)

    # plt.title("Reference Signal (Error)")
    # plt.legend()
    # plt.savefig(os.path.join(sampler.results_dir, 'diff_mean-ref'))

    if charge_device:
        plt.figure()
        plt.plot(deviation_chargeup, label='Device', alpha=0.5)
        if model is not None:
            plt.plot(model_deviation_chargeup,
                     label='Device vs Model',
                     alpha=0.5)
        plt.title("Charging up signal")
        plt.xlabel("Batch number of signal")
        plt.ylabel("RMSE Current (nA)")
        plt.legend()
        plt.savefig(
            os.path.join(sampler.results_dir, 'deviations_while_charging_up'))
    if show_plots:
        plt.show()


# if __name__ == '__main__':
#     import torch
#     from brainspy.processors.processor import Processor
#     from brainspy.utils.pytorch import TorchUtils

#     configs = {}
#     configs["processor_type"] = "simulation"
#     # configs["input_indices"] = [2, 3]
#     configs["electrode_effects"] = {}
#     # configs["electrode_effects"]["amplification"] = [1]
#     # configs["electrode_effects"]["output_clipping"] = [-114, 114]
#     # configs["electrode_effects"]["noise"] = {}
#     # configs["electrode_effects"]["noise"]["noise_type"] = "gaussian"
#     # configs["electrode_effects"]["noise"]["variance"] = 0.6533523201942444
#     configs["driver"] = {}
#     configs["waveform"] = {}
#     configs["waveform"]["plateau_length"] = 1
#     configs["waveform"]["slope_length"] = 0

#     # model_data = torch.load(
#     #     'C:/Users/Unai/Documents/programming/smg/tmp/output/conv_model/training_data_2021_07_19_143513/training_data.pt'
#     # )

#     configs = {}
#     configs['processor_type'] = 'simulation'
#     configs["waveform"] = {}
#     configs["waveform"]["plateau_length"] = 1 #10
#     configs["waveform"]["slope_length"] = 0 #30

#     model_data = {}
#     model_data["info"] = {}
#     model_data["info"]["model_structure"] = {
#         "hidden_sizes": [90, 90, 90],
#         "D_in": 7,
#         "D_out": 1,
#         "activation": "relu",
#     }
#     model_data["info"]['electrode_info'] = {}
#     model_data["info"]['electrode_info']['electrode_no'] = 8
#     model_data["info"]['electrode_info']['activation_electrodes'] = {}
#     model_data["info"]['electrode_info']['activation_electrodes']['electrode_no'] = 7
#     model_data["info"]['electrode_info']['activation_electrodes'][
#             'voltage_ranges'] = np.array([[-0.55, 0.325], [-0.95, 0.55],
#                                           [-1., 0.6], [-1., 0.6], [-1., 0.6],
#                                           [-0.95, 0.55], [-0.55, 0.325]])
#     model_data["info"]['electrode_info']['output_electrodes'] = {}
#     model_data["info"]['electrode_info']['output_electrodes']['electrode_no'] = 1
#     model_data["info"]['electrode_info']['output_electrodes']['amplification'] = [28.5]
#     model_data["info"]['electrode_info']['output_electrodes']['clipping_value'] = None

#     model = Processor(configs, model_data['info'])#,
#                       #model_data['model_state_dict'])
#     model = TorchUtils.format(model)
#     consistency_check(
#         'tests/data/',
#         repetitions=1,
#         model=model)



"""
Automatised generation of deep neural-network based surrogate models of multi-terminal nano-electronic devices, particularly it is focused on dopant-network processing units (DNPUs) [1]. The library provides support for the whole life-cycle of a surrogate model:

-Device preparation: Depending on its properties, such as the material type, devices can exhibit different behaviours. Before the training step, there is a device preparation step, where the IV curves for different selected activation electrodes are studied in different conditions. This library helps visualising the IV curves not only for a single device, but for multiple devices at the same time, enabling to study them simultaneously.

-Data acquisition: The data are sampled using sinusoidal or triangular modulated input signals to minimize capacitive transient effects. The frequencies, amplitudes and offsets of these functions for each activation electrode are given by the voltage ranges chosen during the IV curve observation stage. Frequencies are proportional to the square root of prime numbers. Since the ratio of any two frequencies is irrational, it guarantees a good coverage of the multi-dimensional input voltage space and prevents any recurrence of voltage combinations. In order to have an even quicker coverage, the phase of inputs signals can be randomly shifted. More information can be found in [3].

-Model training: A model is trained using a Pytorch based neural network. The user can customise the size of the neural network as well as the activation functions used for this model. The tool produces error plots of the trained surrogate model on all training, validation and test datasets. It also saves relevant information about how the data was acquired, how the model was trained, and other relevant electrode information that is required for using surrogate models in brains-py.

-Model maintenance: It could be that devices change its behaviour after receiving abrupt changes in temperature or high voltages. In order to check if a model is still behaving as the original hardware, some consistency checks are provided, that help understand the differences between the actual signal of the device, and the signal from the surrogate model, compared to the original data that was gathered for training the device.
"""
TEST_MODE = "SIMULATION_PC"


"""
File containing a class for measuring IV curves on multiple devices PCBs.
"""
import numpy as np
from brainspy.utils.io import load_configs
from brainspy.utils.manager import get_driver
from bspysmg.utils.plots import iv_plot
from bspysmg.utils.inputs import generate_sawtooth_simple

from brainspy.processors.processor import Processor
from brainspy.utils.pytorch import TorchUtils
import torch
import collections


class IVMeasurement():
    
    def __init__(self, configs: dict, info: dict = None,
        model_state_dict: collections.OrderedDict = None) -> None:
        """
        Initializes the driver for which IV curve is to be plotted. It uses a config dict to
        initialize the driver. The driver can be the DNPU device itself (on chip training) or
        a surrogate model (off chip training). This class only allows for measurement of IV
        curves of one device at a time.

        Parameters
        ----------
        configs : dict
            ivcurve configurations with the following keys:
            1. processor_type:  str
            The type of driver for which ivcurve is to be measured. It can take following
            values: cdaq_to_cdaq, cdaq_to_nidaq, simulation_debug.
        """
        self.driver = TorchUtils.format(Processor(configs, info, model_state_dict))

    def iv_curve(self,
                 vmax: float,
                 vmin: float,
                 point_no: int,
                 input_electrode: int,
                 up_direction: bool = True,
                 close: bool = True,
                 show_plot: bool = False,
                 save_plot: bool = None) -> np.array:
        """
        Computes the IV response of DNPU device or surrogate model to an input sawtooth signal.
        Optionally shows the graph on screen and saves it to current directory. This is done
        to check if a particular electrode has non linear IV response or a negative differential
        resistance (NDR).
        Also check:
        
        https://en.wikibooks.org/wiki/Circuit_Idea/Negative_Differential_Resistance
        
        https://resources.pcb.cadence.com/blog/2019-what-is-linear-and-nonlinear-resistance

        Parameters
        ----------
        v_max : float
            Maximum voltage that the sawtooth will achieve.
        v_min : float
            Minimum voltage that the sawtooth will achieve.
        point_no : int
            Number of points that the sawtooth will have.
        input_electrode : int
            Electrode number.
        up_direction : bool [Optional]
            Direction of the sawtooth. If true, the sawtooth will go first up and then down.
            If False, the sawtooth will go first down and then up. By default False.
        close : bool [Optional]
            If set to true, it closes the driver and clears it from memory.
        show_plot : bool [Optional]
            If set to true, it displays the generated plot.
        save_plot : bool [Optional]
            If set to true, it saves the generated plot to current directory.
        
        Returns
        ----------
        result : np.array
            IV response of device or surrogate model.
        """
        data = np.zeros((self.driver.get_activation_electrode_no(), point_no))
        data[input_electrode] = generate_sawtooth_simple(
            vmax, vmin, point_no, up_direction)
        result = self.driver(TorchUtils.format(data.T))
        if close:
            self.driver.close()
        return data[input_electrode], result.detach().cpu().numpy()


"""
File containing a class for sampling a device.
"""
from __future__ import generator_stop
from brainspy.utils.manager import get_driver
from bspysmg.utils.inputs import get_input_generator
from brainspy.utils.io import create_directory_timestamp as mkdir
from brainspy.utils.io import save_configs
from bspysmg.utils.plots import plot_waves
from bspysmg.utils.inputs import get_random_phase
from more_itertools import grouper
import numpy as np
import time
import os
from typing import Tuple, Generator


class Sampler:
    def __init__(self, configs: dict) -> None:
        """
        Initialises the driver from which the data will be acquired, and it stores the sampling
        configurations internally. It also calculates the amplitude and vertical offset of the input
        waveform from the control voltage ranges of the  driver, if the 'amplitude' and 'offset'
        keys are not present in the configuration dictionary.

        Parameters
        ----------
        configs : dict
            Sampling configurations, the dictionary has the following keys:
            1. save_directory: str
            Directory where the all the sampling data will be stored.

            2. data_name: str
            Inside the path specified on the variable save_directory, a folder will be created,
            with the format: <data_name>+<current_timestamp>. This variable specified the
            prefix of that folder before the timestamp.

            3. driver: dict
            Dictionary containing the driver configurations. For more information check the
            documentation about this configuration file, check the documentation of
            brainspy.processors.hardware.drivers.ni.setup.NationalInstrumentsSetup

            4. input_data : dict
            Dictionary containing the information necessary to create the input sampling data.
            4.1 input_distribution: str
            It determines the wave shape of the input. Two main options availeble 'sawtooth'
            and 'sine'. The first option will create saw-like signals, and the second
            sine-wave signals. Sawtooth signals have more coverage on the edges of the
            input range.

            4.2 activation_electrode_no: int
            Number of activation electrodes in the device that wants to be sampled.

            4.3 readout_electrode_no : int
            Number of readout electrodes in the device that wants to be sampled.

            4.4 input_frequency: list
            Base frequencies of the input waves that will be created. In order to optimise
            coverage, irrational numbers are recommended. The list should have the same
            length as the activation electrode number. E.g., for 7 activation electrodes:
            input_frequency = [2, 3, 5, 7, 13, 17, 19]

            4.5 random_phase_shift_each : int
            The input data for each activation electrode can be shifted horizontally by
            changing its phase. This is a randomised process. This variable represents
            at how many batched samples the phase will be randomised. This will cover the
            input space faster and increase the data quality such that less data is needed.
            If the value is 0, no phase shift will be applied.

            4.6 factor : float
            Given factor by which the input frequencies will be multiplied after square
            rooting them.

            4.7 amplitude : Optional[list[float]]
            Amplitude of the generated input wave signal. It is calculated according to the
            minimum and maximum ranges of each electrode. Where the amplitude value should
            correspond with (max_range_value - min_range_value) / 2. If no amplitude is
            given it will be automatically calculated from the driver configurations for
            activation electrode ranges. If it wants to be manually set, the offset
            variable should also be included in the dictionary.

            4.8 offset: Optional[list[float]]
            Vertical offset of the generated input wave signal. It is calculated according
            to the minimum and maximum ranges of each electrode. Where the offset value
            should correspond with (max_range_value + min_range_value) / 2. If no offset
            is given it will be automatically calculated from the driver configurations for
            activation electrode ranges. If it wants to be manually set, the offset
            variable should also be included in the dictionary.

            4.9 ramp_time: float
            Time that will be taken before sending each batch to go from zero to the first
            point of the batch and to zero from the last point of the batch.

            4.10  batch_time:
            Time that the sampling of each batch will take.

            4.11 number_batches: int
            Number of batches that will be sampled. A default value of 3880 is reccommended.

            4.12 randomise_phase_each: int (Optional)
            Specifies at how many epochs the phases will be randomly generated again. 
        """
        self.driver = get_driver(configs["driver"])
        self.configs = configs
        if 'amplitude' not in self.configs[
                'input_data'] and 'offset' not in self.configs['input_data']:
            self.config_offset_and_amplitude()

    def config_offset_and_amplitude(self) -> None:
        """
        It extracts the offset and amplitude values that the input waveforms will have, according
        to the voltage ranges specified in the driver. It stores them into the configuration
        dictionary (input_data/amplitude and input_data/offset).
        """
        assert isinstance(
            self.configs['driver']['instruments_setup']
            ['activation_voltage_ranges'],
            list), "Voltage ranges should be passed as a list"
        assert self.configs['driver']['instruments_setup'][
            'activation_voltage_ranges'] != [], "Empty array for voltage ranges"

        voltage_ranges = np.array(self.configs['driver']['instruments_setup']
                                  ['activation_voltage_ranges'])
        amplitude = (voltage_ranges[:, 1] - voltage_ranges[:, 0]) / 2
        offset = (voltage_ranges[:, 0] + voltage_ranges[:, 1]) / 2
        self.configs['input_data']['amplitude'] = amplitude
        self.configs['input_data']['offset'] = offset
        print(f"Amplitude: {str(amplitude)}")
        print(f"Offset: {str(offset)}")

    def sample_batch(self, x: np.array) -> np.array:
        """
        Perform a sampling operation over one input batch. This method includes
        a ramping from zero until the beginning of the batch until the first point, and
        a ramping from the last point to zero. The information from ramps is filtered
        before it is returned.

        Parameters
        ----------
        x : np.array
            Input batch that will be sent to the device in order to sample its corresponding output.
            The dimension of the sample should be (activation_electrode_no, batch_size).

        Returns
        -------
        output : np.array
            Readout of the device when applying the input batch. The size of the output will be
            (batch_size, readout_electrode_no).
        """
        # Ramp input batch (0.5 sec up and down)
        ramped_input = self.ramp_input(x)
        outputs_ramped = self.driver.forward_numpy(ramped_input.T)
        return outputs_ramped[self.filter_ramp]

    def ramp_input(self, x: np.array):
        """
        The input batch is prepared for sampling on the device by ramping it from zero until the
        beginning of the batch until the first point, and a ramping from the last point to zero.

        Parameters
        ----------
        x : np.array
            Input batch that will be sent to the device in order to sample its corresponding output.
            The dimension of the sample should be (activation_electrode_no, batch_size). 

        Returns
        -------
        ramped_input : np.array
            Input batch that will be sent to the device in order to sample its corresponding output
            with an additional ramping from zero until the beginning of the batch until the first
            point, and a ramping from the last point to zero.
        """
        dimensions = self.configs["input_data"]["activation_electrode_no"]
        ramp = int(self.configs["input_data"]["ramp_points"])
        ramped_input = np.zeros((dimensions, self.nr_points_ramped_signal))
        for j in range(dimensions):
            ramped_input[j, 0:ramp] = np.linspace(0, x[j, 0], ramp)
            ramped_input[j, ramp:self.end_batch] = x[j, :]
            ramped_input[j, self.end_batch:] = np.linspace(x[j, -1], 0, ramp)
        return ramped_input

    def sample(self, plot_interval: int = 1) -> str:
        """
        Performs a full sampling operation, divided into several batches,
        according to the configurations given to the class. It stores each
        batch on a txt file. Additionally it can also show the inputs and
        corresponding outputs of a batch in a saved plot.

        Parameters
        ----------
        plot_interval : Optional[int]
            It sets after how many batches it will save a plot of the current batch,
            that shows the inputs and outputs. By default 1 (on every batch).

        Returns
        -------
        save_directory : str
            Folder where all the sampling data is stored.
        """
        # Create a directory and file for saving
        self.save_batch()
        # Initialize configs
        total_number_samples, batch_size, input_dict = self.init_configs()

        # Initialize sampling loop
        all_time_points = np.arange(
            total_number_samples) / input_dict["sampling_frequency"]
        if 'phase' in self.configs['input_data']:
            phase = np.array(self.configs['input_data']['phase'])
            if len(phase.shape) == 1:
                phase = phase[:,np.newaxis]
        else:
            phase = np.zeros(
                (self.configs["input_data"]["activation_electrode_no"], 1))
        phase_randomisation_count = 0
        for batch, batch_indices in enumerate(
                self.get_batch_indices(total_number_samples, batch_size)):
            start_batch = time.time()

            # Generate inputs (without ramping)
            batch += 1
            time_points = all_time_points[batch_indices]
            inputs = self.generate_inputs(time_points,
                                          input_dict['input_frequency_numpy'], phase,
                                          input_dict['amplitude_numpy'],
                                          input_dict['offset_numpy'])
            # Get outputs (without ramping)
            outputs = self.sample_batch(inputs)
            self.save_batch(inputs.T, outputs)
            end_batch = time.time()
            if (batch % plot_interval) == 0:
                input_no = self.configs["input_data"][
                    "activation_electrode_no"]
                output_no = self.configs["input_data"]["readout_electrode_no"]
                legend = self.get_header(input_no, output_no).split(',')
                plot_waves(inputs.T, outputs, input_no, output_no, batch,
                           legend, self.configs["save_directory"])
            print(
                f"Outputs collection for batch {batch} of {input_dict['number_batches']} "
                + f"took {end_batch - start_batch} sec. Estimated time left: "+convert(batch, input_dict['number_batches'], (end_batch - start_batch)) )
            phase_randomisation_count += 1
            if input_dict[
                    'random_phase_shift_each'] >= phase_randomisation_count:
                phase = get_random_phase(
                    self.configs["input_data"]["activation_electrode_no"])
                phase_randomisation_count = 0
        self.close_driver()
        return self.configs["save_directory"]

    def get_batch_indices(self, sample_no: int,
                          batch_size: int) -> Generator[int, None, None]:
        """
        Collects data length into indices and yields them into fixed-length chunks or blocks.

        Parameters
        ----------
        sample_no : int
            Total number of samples to be sent to the device.
        batch_size : int
            Desired block size in which the total number of sizes will be divided.

        Yields
        -------
        indices : list[int]
            List of indices corresponding to the generated input signal.

        """
        print('Start batching...')
        batches = grouper(np.arange(sample_no), batch_size)
        while True:
            try:
                indices = list(next(batches))
                if None in indices:
                    indices = [index for index in indices if index is not None]
                yield indices
            except StopIteration:
                return

    def get_header(self, input_no: int, output_no: int) -> str:
        """
        Gets the header of the txt file data, so that is stored as a string format.

        Parameters
        ----------
        input_no : int
            The input electrode number.
        output_no : int
            The output electrode number.

        Returns
        -------
        str
            The headers of each input and output batch in a string format.
        """
        header = ""
        for i in range(input_no):
            header += f"Input {i}, "
        for i in range(output_no):
            if i < (output_no - 1):
                header += f"Output {i}, "
            else:
                header += f"Output {i}"
        return header

    def init_configs(self) -> Tuple[int, int, dict]:
        """
        Initializes the configurations for performing sampling operation. It initializes
        and returns the number of samples, the batch size and an input dictionary for
        sampling.

        Returns
        -------
        tuple
            total_number_samples: int
                The total number of samples to be generated in the input dataset.
            batch_size: int
                The batch size to be used for processing the input dataset in batches.
            input_dict: dict
                The configurations for sampling operation with following keys:
                    - activation_electrode_no: int
                        Number of activation electrodes in the device that wants to be sampled.
                    - readout_electrode_no : int
                        Number of readout electrodes in the device that wants to be sampled.
                    - input_frequency: list
                        Base frequencies of the input waves that will be created. In order to optimise
                        coverage, irrational numbers are recommended. The list should have the same
                        length as the activation electrode number. E.g., for 7 activation electrodes:
                        input_frequency = [2, 3, 5, 7, 13, 17, 19]
                    - phase : float
                        Horizontal shift of the input signals. It is recommended to have random numbers
                        which are different for the training, validation and test datasets. These
                        numbers will be square rooted and multiplied by a given factor.
                    - amplitude : Optional[list[float]]
                        Amplitude of the generated input wave signal. It is calculated according to the
                        minimum and maximum ranges of each electrode. Where the amplitude value should
                        correspond with (max_range_value - min_range_value) / 2. If no amplitude is
                        given it will be automatically calculated from the driver configurations for
                        activation electrode ranges. If it wants to be manually set, the offset
                        variable should also be included in the dictionary.
                    - offset: Optional[list[float]]
                        Vertical offset of the generated input wave signal. It is calculated according
                        to the minimum and maximum ranges of each electrode. Where the offset value
                        should correspond with (max_range_value + min_range_value) / 2. If no offset
                        is given it will be automatically calculated from the driver configurations for
                        activation electrode ranges. If it wants to be manually set, the offset
                        variable should also be included in the dictionary.
                    - number_batches: int
                        Number of batches that will be sampled. A default value of 3880 is reccommended.
        """
        input_dict, self.generate_inputs = get_input_generator(self.configs)

        batch_size = input_dict["sampling_frequency"] * input_dict["batch_time"]
        total_number_samples = input_dict["number_batches"] * batch_size

        # define internal attributes
        self.end_batch = int(input_dict["ramp_points"] +
                             input_dict["batch_points"])
        self.nr_points_ramped_signal = int(input_dict["batch_points"] +
                                           2 * input_dict["ramp_points"])

        self.filter_ramp = np.zeros(self.nr_points_ramped_signal, dtype=bool)
        self.filter_ramp[int(input_dict["ramp_points"]):int(self.end_batch
                                                            )] = True

        return int(total_number_samples), int(batch_size), input_dict

    def save_batch(self, *args) -> None:
        """
        Stores a batch of data on the same IO.dat text comma separated values file.
        
        Parameters
        ----------
        *args: Tuple[np.array]
            A set of input and output arrays of a batch which is to be saved.
        """
        if len(args) > 0:
            with open(self.path_to_iodata, '+a') as f:
                data = np.column_stack(args)
                np.savetxt(f, data)
                f.flush()
        else:
            print(f'Saving in {self.configs["save_directory"]}')
            path_to_file = mkdir(self.configs["save_directory"],
                                 self.configs["data_name"])
            self.configs["save_directory"] = path_to_file
            save_configs(self.configs,
                         os.path.join(path_to_file, 'sampler_configs.json'))
            header = self.get_header(
                self.configs["input_data"]["activation_electrode_no"],
                self.configs["input_data"]["readout_electrode_no"])
            self.path_to_iodata = path_to_file + "/IO.dat"
            with open(self.path_to_iodata, 'wb') as f:
                np.savetxt(f, [], header=header)

    def close_driver(self) -> None:
        """
        Adequately closes the connection to the drivers.
        """
        self.driver.close_tasks()
        print('Instrument task closed')

def convert(num_batches, total_batches, time_taken):
    """
    Converts the number of batches and the last time taken for measuring them into
    a string containing an estimation of the time left in HH:MM:SS format.

    Parameters
    ----------
    num_batches: int
        Number of batches that have been already sampled.
    total_batches:int
        Total batches that will be sampled.
    time_taken: float
        Time that the last measurement has taken.
    
    Return
    ----------
        str: A string containing the time left in HH:MM:SS
    """
    seconds = (total_batches - num_batches) * time_taken
    min, sec = divmod(seconds, 60)
    hour, min = divmod(min, 60)
    return "%d:%02d:%02d" % (hour, min, sec)



import unittest
from bspysmg.utils.iv import simple
import numpy as np
from bspysmg import TEST_MODE
from brainspy.utils.io import load_configs

class Test_Simple(unittest.TestCase):
    def __init__(self, *args, **kwargs) -> None:
        super(Test_Simple, self).__init__(*args, **kwargs)
        self.configs = load_configs("configs/utils/brains_ivcurve_template_simple.yaml")

    @unittest.skipUnless(TEST_MODE == "HARDWARE_CDAQ"
                         or TEST_MODE == "HARDWARE_NIDAQ",
                         "Hardware test is skipped for simulation setup.")
    def test_init(self):

        try:
            measurement = simple.IVMeasurement(self.configs)
            measurement.iv_curve(
             measurement.driver.get_voltage_ranges()[0, 0].item(),
             measurement.driver.get_voltage_ranges()[0, 1].item(),
             point_no=10,
             input_electrode=0,
             show_plot=True,
             close=False)
            measurement.iv_curve(
             measurement.driver.get_voltage_ranges()[1, 0].item(),
             measurement.driver.get_voltage_ranges()[1, 1].item(),
             point_no=10,
             input_electrode=1,
             show_plot=True,
             close=True)
        except Exception:
            self.fail("Failed in creating error hist")


if __name__ == '__main__':
    import bspysmg

    bspysmg.TEST_MODE = 'HARDWARE_CDAQ'
    unittest.main()



"""
File containing functions for training a surrogate model in pytorch taking into account the error in nano amperes.
"""
import os
import torch
import matplotlib.pyplot as plt

import numpy as np

# from brainspy.algorithm_manager import get_algorithm
# from bspysmg.model.data.inputs.data_handler import get_training_data
from tqdm import tqdm

from torch.optim import Adam
from torch.nn import MSELoss

from brainspy.utils.pytorch import TorchUtils
from brainspy.utils.io import create_directory_timestamp
from brainspy.processors.simulation.model import NeuralNetworkModel
from bspysmg.data.dataset import get_dataloaders
from bspysmg.utils.plots import plot_error_vs_output, plot_error_hist
from typing import Tuple, List


def init_seed(configs: dict) -> None:
    """
    Initializes a random seed for training. A random seed is a starting point for pseudorandom
    number generator algorithms which is used for reproducibility.
    Also see - https://pytorch.org/docs/stable/notes/randomness.html  

    Parameters
    ----------
    configs : dict
        Training configurations with the following keys:

        1. seed:  int [Optional]
        The desired seed for the random number generator. If the dictionary does not contain 
        this key, a deterministic random seed will be applied, and added to the key 'seed' in 
        the dictionary.
    """
    if "seed" in configs:
        seed = configs["seed"]
    else:
        seed = None

    seed = TorchUtils.init_seed(seed, deterministic=True)
    configs["seed"] = seed


def generate_surrogate_model(
        configs: dict,
        custom_model: torch.nn.Module = NeuralNetworkModel,
        criterion: torch.nn.modules.loss._Loss = MSELoss(),
        custom_optimizer: torch.optim.Optimizer = Adam,
        main_folder: str = "training_data") -> None:
    """
    It loads the training and validation datasets from the npz file specified
    in the field data/dataset_paths of the configs dictionary. These npz files
    can be created when running the postprocessing of the sampling data. The
    method will train a neural network with the structure specified in the
    model_structure field of the configs using the loaded training and validation
    datasets. It provides, on the specified saving directory, a trained model,
    the plots of the training performance, and the error of the model.

    Note that the method will only save a model (in each epoch) if current validation
    loss is less than the validation loss from the previous epoch.

    Parameters
    ----------
    configs : dict
        Training configurations for training a model with following keys:
        
        1. results_base_dir: str
        Directory where the trained model and corresponding performance plots will be stored.

        2. seed: int
        Sets the seed for generating random numbers to a non-deterministic random number.

        3. hyperparameters:
        epochs: int
        learning_rate: float
        
        4. model_structure: dict
        The definition of the internal structure of the surrogate model, which is typically five
        fully-connected layers of 90 nodes each.

        4.1 hidden_sizes : list
        A list containing the number of nodes of each layer of the surrogate model.
        E.g., [90,90,90,90,90]

        4.2 D_in: int
        Number of input features of the surrogate model structure. It should correspond to
        the activation electrode number.

        4.3 D_out: int
        Number of output features of the surrogate model structure. It should correspond to
        the readout electrode number.

        5. data:
        5.1 dataset_paths: list[str]
        A list of paths to the Training, Validation and Test datasets, stored as
        postprocessed_data.npz

        5.2 steps : int
        It allows to skip parts of the data when loading it into memory. The number indicates
        how many items will be skipped in between. By default, step number is one (no values
        are skipped). E.g., if steps = 2, and the inputs are [0, 1, 2, 3, 4, 5, 6]. The only
        inputs taken into account would be: [0, 2, 4, 6].

        5.3 batch_size: int
        How many samples will contain each forward pass.

        5.4 worker_no: int
        How many subprocesses to use for data loading. 0 means that the data will be loaded in
        the main process. (default: 0)

        5.5 pin_memory: boolean
        If True, the data loader will copy Tensors into CUDA pinned memory before returning
        them. If your data elements are a custom type, or your collate_fn returns a batch that
        is a custom type.
        custom_model : custom model of type torch.nn.Module
            Model to be trained.
        criterion : <method>
            Loss function that will be used to train the model.
        custom_optimizer : torch.optim.Optimizer
            Optimization method used to train the model which decreases model's loss.
        save_dir : string [Optional]
            Name of the path where the trained model is to be saved.
    
    Return
    ------
    saved_dir: str 
        Directory where the surrogate model was saved.
    """
    # Initialise seed and create data directories
    init_seed(configs)
    results_dir = create_directory_timestamp(configs["results_base_dir"],
                                             main_folder)

    # Get training, validation and test data
    # Get amplification of the device and the info
    dataloaders, amplification, info_dict = get_dataloaders(configs)

    # Initilialise model
    model = custom_model(info_dict["model_structure"])
    # model.set_info_dict(info_dict)
    model = TorchUtils.format(model)

    # Initialise optimiser
    optimizer = custom_optimizer(
        filter(lambda p: p.requires_grad, model.parameters()),
        lr=configs["hyperparameters"]["learning_rate"],
        betas=(0.9, 0.75),
    )

    # Whole training loop
    model, performances, saved_dir = train_loop(
        model,
        info_dict,
        (dataloaders[0], dataloaders[1]),
        criterion,
        optimizer,
        configs["hyperparameters"]["epochs"],
        amplification,
        save_dir=results_dir,
    )

    # Plot results
    labels = ["TRAINING", "VALIDATION", "TEST"]
    for i in range(len(dataloaders)):
        if dataloaders[i] is not None:
            loss = postprocess(
                dataloaders[i],
                model,
                criterion,
                amplification,
                results_dir,
                label=labels[i],
            )

    plt.figure()
    plt.plot(TorchUtils.to_numpy(performances[0]))
    if len(performances) > 1 and not len(performances[1]) == 0:
        plt.plot(TorchUtils.to_numpy(performances[1]))
    if dataloaders[-1].tag == 'test':
        plt.plot(np.ones(len(performances[-1])) * TorchUtils.to_numpy(loss))
        plt.title("Training profile /n Test loss : %.6f (nA)" % loss)
    else:
        plt.title("Training profile")
    if not len(performances[1]) == 0:
        plt.legend(["training", "validation"])
    plt.xlabel("Epoch no.")
    plt.ylabel("RMSE (nA)")
    plt.savefig(os.path.join(results_dir, "training_profile"))
    if not dataloaders[-1].tag == 'train':
        training_data = torch.load(
            os.path.join(results_dir, "training_data.pt"))
        training_data['test_loss'] = loss
        torch.save(training_data, os.path.join(results_dir,
                                               "training_data.pt"))
    # print("Model saved in :" + results_dir)
    return saved_dir


def train_loop(
    model: torch.nn.Module,
    info_dict: dict,
    dataloaders: List[torch.utils.data.DataLoader],
    criterion: torch.nn.modules.loss._Loss,
    optimizer: torch.optim.Optimizer,
    epochs: int,
    amplification: float,
    start_epoch: int = 0,
    save_dir: str = None,
    early_stopping: bool = True,
) -> Tuple[torch.nn.Module, List[float]]:
    """
    Performs the training of a model and returns the trained model, training loss
    validation loss. It also saves the model in each epoch if current validation
    loss is less than the previous validation loss.

    Parameters
    ----------
    model : custom model of type torch.nn.Module
        Model to be trained.
    info_dict : dict
        The dictionary used for initialising the surrogate model. It has the following keys:
        1. model_structure: dict
        The definition of the internal structure of the surrogate model, which is typically five
        fully-connected layers of 90 nodes each.

        1.1 hidden_sizes : list
        A list containing the number of nodes of each layer of the surrogate model.
        E.g., [90,90,90,90,90]

        1.2 D_in: int
        Number of input features of the surrogate model structure. It should correspond to
        the activation electrode number.

        1.3 D_out: int
        Number of output features of the surrogate model structure. It should correspond to
        the readout electrode number.

        2. electrode_info: dict
        It contains all the information required for the surrogate model about the electrodes.
        2.1 electrode_no: int
        Total number of electrodes in the device

        2.2 activation_electrodes: dict
        2.2.1 electrode_no: int
        Number of activation electrodes used for gathering the data
        
        2.2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and
        maximum of the ranges, respectively.

        2.3 output_electrodes: dict

        2.3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        2.3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified
        values.

        2.3.3 amplification: float
        Amplification correction factor used in the device to correct the
        amplification applied to the output current in order to convert it into
        voltage before its readout.

        3. training_configs: dict
        A copy of the configurations used for training the surrogate model.

        4. sampling_configs : dict
        A copy of the configurations used for gathering the training data.
    dataloaders :  list
        A list containing a single PyTorch Dataloader containing the training dataset.
    criterion : <method>
        Loss function that will be used to train the model.
    optimizer : torch.optim.Optimizer
        Optimization method used to train the model which decreases model's loss.
    epochs : int
        The number of iterations for which the model is to be trained.
    amplification: float
        Amplification correction factor used in the device to correct the amplification
        applied to the output current in order to convert it into voltage before its
        readout.
    start_epoch : int [Optional]
        The starting value of the epochs.
    save_dir : string [Optional]
        Name of the path and file where the trained model is to be saved.
    early_stopping : bool [Optional]
        If this is set to true, early stopping algorithm is used during the training
        of the model.
        Also see - https://medium.com/analytics-vidhya/early-stopping-with-pytorch-to-
        restrain-your-model-from-overfitting-dce6de4081c5

    Returns
    -------
    model
        Trained model
    losses
        list of training loss and validation loss.
    saved_dir
        directory where the model was saved
    """
    if start_epoch > 0:
        start_epoch += 1

    train_losses, val_losses = TorchUtils.format([]), TorchUtils.format([])
    min_val_loss = np.inf

    for epoch in range(epochs):
        print("\nEpoch: " + str(epoch))
        model, running_loss = default_train_step(model, dataloaders[0],
                                                 criterion, optimizer)
        running_loss = running_loss**(1 / 2)
        running_loss *= amplification
        train_losses = torch.cat((train_losses, running_loss.unsqueeze(dim=0)),
                                 dim=0)
        description = "Training loss (RMSE): {:.6f} (nA)\n".format(
            train_losses[-1].item())

        if dataloaders[1] is not None and len(dataloaders[1]) > 0:
            val_loss = default_val_step(model, dataloaders[1], criterion)
            val_loss = val_loss**(1 / 2)
            val_loss *= amplification
            val_losses = torch.cat((val_losses, val_loss.unsqueeze(dim=0)),
                                   dim=0)
            description += "Validation loss (RMSE): {:.6f} (nA)\n".format(
                val_losses[-1].item())
            # Save only when peak val performance is reached
            if (save_dir is not None and early_stopping
                    and val_losses[-1] < min_val_loss):
                min_val_loss = val_losses[-1]
                description += "Model saved in: " + save_dir
                # torch.save(model, os.path.join(save_dir, "model.pt"))
                torch.save(
                    {
                        "epoch": epoch,
                        "model_state_dict": model.state_dict(),
                        "info": info_dict,
                        "optimizer_state_dict": optimizer.state_dict(),
                        "train_losses": train_losses,
                        "val_losses": val_losses,
                        "min_val_loss": min_val_loss,
                    },
                    os.path.join(save_dir, "training_data.pt"),
                )

        print(description)
        # looper.set_description(description)

    # TODO: Add a save instruction and a stopping criteria
    # if stopping_criteria(train_losses, val_losses):
    #     break
    print("\nFinished training model. ")
    print("Model saved in: " + save_dir)
    if (save_dir is not None and early_stopping and dataloaders[1] is not None
            and len(dataloaders[1]) > 0):
        training_data = torch.load(os.path.join(save_dir, "training_data.pt"))
        model.load_state_dict(training_data["model_state_dict"])
        print("Min validation loss (RMSE): {:.6f} (nA)\n".format(
            min_val_loss.item()))
    else:
        torch.save(
            {
                "epoch": epoch,
                "model_state_dict": model.state_dict(),
                "info": info_dict,
                "optimizer_state_dict": optimizer.state_dict(),
                "train_losses": train_losses,
                "val_losses": val_losses,
                "min_val_loss": min_val_loss,
            },
            os.path.join(save_dir, "training_data.pt"),
        )

    return model, [train_losses, val_losses], save_dir


def default_train_step(
        model: torch.nn.Module, dataloader: torch.utils.data.DataLoader,
        criterion: torch.nn.modules.loss._Loss,
        optimizer: torch.optim.Optimizer) -> Tuple[torch.nn.Module, float]:
    """
    Performs the training step of a model within a single epoch and returns the
    current loss and current trained model.

    Parameters
    ----------
    model : custom model of type torch.nn.Module
        Model to be trained.
    dataloader :  torch.utils.data.DataLoader
        A PyTorch Dataloader containing the training dataset.
    criterion : <method>
        Loss function that will be used to train the model.
    optimizer : torch.optim.Optimizer
        Optimization method used to train the model which decreases model's loss.

    Returns
    -------
    tuple
        Trained model and training loss for the current epoch.
    """
    running_loss = 0
    model.train()
    loop = tqdm(dataloader)
    for inputs, targets in loop:
        inputs, targets = to_device(inputs), to_device(targets)
        optimizer.zero_grad()
        predictions = model(inputs)
        loss = criterion(predictions, targets)
        loss.backward()
        optimizer.step()
        running_loss += loss.item() * inputs.shape[0]
        loop.set_postfix(batch_loss=loss.item())
    running_loss /= len(dataloader.dataset)
    return model, running_loss


def default_val_step(model: torch.nn.Module,
                     dataloader: torch.utils.data.DataLoader,
                     criterion: torch.nn.modules.loss._Loss) -> float:
    """
    Performs the validation step of a model within a single epoch and returns
    the validation loss.

    Parameters
    ----------
    model : custom model of type torch.nn.Module
        Model to be trained.
    dataloader :  torch.utils.data.DataLoader
        A PyTorch Dataloader containing the training dataset.
    criterion : <method>
        Loss function that will be used to train the model.

    Returns
    -------
    float
        Validation loss for the current epoch.
    """
    with torch.no_grad():
        val_loss = 0
        model.eval()
        loop = tqdm(dataloader)
        for inputs, targets in loop:
            inputs, targets = to_device(inputs), to_device(targets)
            predictions = model(inputs)
            loss = criterion(predictions, targets)
            val_loss += loss.item() * inputs.shape[0]
            loop.set_postfix(batch_loss=loss.item())
        val_loss /= len(dataloader.dataset)
    return val_loss


def postprocess(dataloader: torch.utils.data.DataLoader,
                model: torch.nn.Module, criterion: torch.nn.modules.loss._Loss,
                amplification: float, results_dir: str, label: str) -> float:
    """
    Plots error vs output and error histogram for given dataset and saves it to
    specified directory.

    Parameters
    ----------
    dataloader :  torch.utils.data.DataLoader
        A PyTorch Dataloader containing the training dataset.
    model : custom model of type torch.nn.Module
        Model to be trained.
    criterion : <method>
        Loss function that will be used to train the model.
    amplification: float
        Amplification correction factor used in the device to correct the amplification
        applied to the output current in order to convert it into voltage before its
        readout.
    results_dir : string
        Name of the path and file where the plots are to be saved.
    label : string
        Name of the dataset. I.e., train, validation or test.

    Returns
    -------
    float
        Mean Squared error evaluated on given dataset.
    """
    print(f"Postprocessing {label} data ... ")
    # i = 0
    running_loss = 0
    all_targets = []
    all_predictions = []
    with torch.no_grad():
        model.eval()
        for inputs, targets in tqdm(dataloader):
            inputs, targets = to_device(inputs), to_device(targets)
            predictions = model(inputs)
            all_targets.append(amplification * targets)
            all_predictions.append(amplification * predictions)
            loss = criterion(predictions, targets)
            running_loss += loss * inputs.shape[0]  # sum up batch loss

    running_loss /= len(dataloader.dataset)
    running_loss = running_loss * (amplification**2)

    print(label.capitalize() +
          " loss (MSE): {:.6f} (nA)".format(running_loss.item()))
    print(
        label.capitalize() +
        " loss (RMSE): {:.6f} (nA)\n".format(torch.sqrt(running_loss).item()))

    all_targets = TorchUtils.to_numpy(torch.cat(all_targets, dim=0))
    all_predictions = TorchUtils.to_numpy(torch.cat(all_predictions, dim=0))

    error = all_targets - all_predictions

    plot_error_vs_output(
        all_targets,
        error,
        results_dir,
        name=label + "_error_vs_output",
    )
    plot_error_hist(
        all_targets,
        all_predictions,
        error,
        TorchUtils.to_numpy(running_loss),
        results_dir,
        name=label + "_error",
    )
    return torch.sqrt(running_loss)


def to_device(inputs: torch.Tensor) -> torch.Tensor:
    """
    Copies input tensors from CPU to GPU device for processing. GPU allows multithreading
    which makes computation faster. The rule of thumb is using 4 worker threads per GPU.
    See - https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device

    Parameters
    ----------
    inputs : torch.Tensor
        Input tensor which needs to be loaded into GPU device.

    Returns
    -------
    torch.Tensor
        Input tensor allocated to GPU device.
    """
    if inputs.device != TorchUtils.get_device():
        inputs = inputs.to(device=TorchUtils.get_device())
    return inputs


import unittest
from bspysmg.utils.iv import multiple
import numpy as np
from bspysmg import TEST_MODE
from brainspy.utils.io import load_configs
from bspysmg.utils import plots

class Test_Multiple(unittest.TestCase):
    def __init__(self, *args, **kwargs) -> None:
        super(Test_Multiple, self).__init__(*args, **kwargs)
        self.configs = load_configs("configs/utils/brains_ivcurve_template.yaml")

    @unittest.skipUnless(TEST_MODE == "HARDWARE_CDAQ"
                         or TEST_MODE == "HARDWARE_NIDAQ",
                         "Hardware test is skipped for simulation setup.")
    def test_init(self):
        try:
            self.configs['driver']['instruments_setup']['A']['activation_channel_mask'] = [1,1,1,1,1,1,1]
            measurement = multiple.MultiIVMeasurement(self.configs)
            configs, inputs, output = measurement.run_test(close_driver=False)
            self.configs['input_signal']['input_signal_type'] = "sine"
            configs, inputs, output = measurement.run_test(close_driver=True)
            self.configs['driver']['instruments_setup']['A']['activation_channel_mask'] = [1,1,0,1,1,1,1]
            measurement = multiple.MultiIVMeasurement(self.configs)
            configs, inputs, output = measurement.run_test(close_driver=True)
            plots.multi_iv_plot(configs, inputs, output, show_plot=False)
        except Exception:
            self.fail("Failed in creating error hist")


if __name__ == '__main__':
    import bspysmg

    bspysmg.TEST_MODE = 'HARDWARE_CDAQ'
    unittest.main()



import unittest
from bspysmg.model import training
from brainspy.utils.io import create_directory_timestamp
from bspysmg.data.dataset import get_dataloaders
from brainspy.processors.simulation.model import NeuralNetworkModel
from brainspy.utils.pytorch import TorchUtils
from torch.optim import Adam
from torch.nn import MSELoss


class Test_Training(unittest.TestCase):
    def __init__(self, *args, **kwargs) -> None:
        super(Test_Training, self).__init__(*args, **kwargs)
        self.configs = {'results_base_dir': 'tests/data'}
        self.configs['model_structure'] = {
            'hidden_sizes': [90] * 5,
            'D_in': 7,
            'D_out': 1,
            'activation': 'relu'
        }
        self.configs['hyperparameters'] = {
            'epochs': 2,
            'learning_rate': 1.0e-05
        }
        self.configs['data'] = {
            'dataset_paths': [
                "tests/data/postprocessed_data.npz",
                "tests/data/postprocessed_data.npz",
                "tests/data/postprocessed_data.npz"
            ],
            'steps':
            4,
            'batch_size':
            128,
            'worker_no':
            0,
            'pin_memory':
            False,
            'split_percentages': [1, 0]
        }

    def test_init_seed(self):
        with self.assertRaises(TypeError):
            training.init_seed()

        try:
            training.init_seed(self.configs)
        except Exception:
            self.fail("Failed Execution: init_seed()")

        with self.assertRaises(TypeError):
            training.init_seed([])

    def test_surrogate_model(self):
        with self.assertRaises(TypeError):
            training.generate_surrogate_model()

        training.generate_surrogate_model(self.configs, main_folder='tests/data')

    def test_train_loop(self):

        try:
            training.init_seed(self.configs)
            results_dir = create_directory_timestamp(
                self.configs["results_base_dir"], 'tests/data')

            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)

            model = NeuralNetworkModel(info_dict["model_structure"])
            model = TorchUtils.format(model)

            optimizer = Adam(
                filter(lambda p: p.requires_grad, model.parameters()),
                lr=self.configs["hyperparameters"]["learning_rate"],
                betas=(0.9, 0.75),
            )

            model, performances,_ = training.train_loop(
                model,
                info_dict, (dataloaders[0], dataloaders[1]),
                MSELoss(),
                optimizer,
                self.configs["hyperparameters"]["epochs"],
                amplification,
                save_dir=results_dir)
        except Exception:
            self.fail("Failed Execution: train_loop()")

    def test_train_loop_no_early_stopping(self):

        try:
            training.init_seed(self.configs)
            results_dir = create_directory_timestamp(
                self.configs["results_base_dir"], 'tests/data')

            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)

            model = NeuralNetworkModel(info_dict["model_structure"])
            model = TorchUtils.format(model)

            optimizer = Adam(
                filter(lambda p: p.requires_grad, model.parameters()),
                lr=self.configs["hyperparameters"]["learning_rate"],
                betas=(0.9, 0.75),
            )

            model, performances, _ = training.train_loop(
                model,
                info_dict, (dataloaders[0], None, None),
                MSELoss(),
                optimizer,
                self.configs["hyperparameters"]["epochs"],
                amplification,
                early_stopping=False,
                save_dir=results_dir)
        except Exception:
            self.fail("Failed Execution: train_loop()")

    def test_train_step(self):

        try:
            training.init_seed(self.configs)

            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)

            model = NeuralNetworkModel(info_dict["model_structure"])
            model = TorchUtils.format(model)

            optimizer = Adam(
                filter(lambda p: p.requires_grad, model.parameters()),
                lr=self.configs["hyperparameters"]["learning_rate"],
                betas=(0.9, 0.75),
            )

            model, running_loss = training.default_train_step(
                model, dataloaders[0], MSELoss(), optimizer)
        except:
            self.fail("Failed Execution: train_step()")

    def test_val_step(self):

        try:
            training.init_seed(self.configs)

            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)

            model = NeuralNetworkModel(info_dict["model_structure"])
            model = TorchUtils.format(model)

            optimizer = Adam(
                filter(lambda p: p.requires_grad, model.parameters()),
                lr=self.configs["hyperparameters"]["learning_rate"],
                betas=(0.9, 0.75),
            )

            val_loss = training.default_val_step(model, dataloaders[1],
                                                 MSELoss())
        except:
            self.fail("Failed Execution: val_step()")

    def test_to_device(self):
        try:
            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)
            for input, target in dataloaders[0]:
                input, target = training.to_device(input), training.to_device(
                    target)
        except:
            self.fail("Failed Execution: to_device()")

    def test_postprocess(self):
        try:
            training.init_seed(self.configs)
            results_dir = create_directory_timestamp(
                self.configs["results_base_dir"], 'tests/data')

            dataloaders, amplification, info_dict = get_dataloaders(
                self.configs)

            model = NeuralNetworkModel(info_dict["model_structure"])
            model = TorchUtils.format(model)

            optimizer = Adam(
                filter(lambda p: p.requires_grad, model.parameters()),
                lr=self.configs["hyperparameters"]["learning_rate"],
                betas=(0.9, 0.75),
            )

            model, performances, _ = training.train_loop(
                model,
                info_dict,
                (dataloaders[0], dataloaders[1]),
                MSELoss(),
                optimizer,
                self.configs["hyperparameters"]["epochs"],
                amplification,
                save_dir=results_dir,
            )
            labels = ["TRAINING", "VALIDATION", "TEST"]
            for i in range(len(dataloaders)):
                if dataloaders[i] is not None:
                    loss = training.postprocess(
                        dataloaders[i],
                        model,
                        MSELoss(),
                        amplification,
                        results_dir,
                        label=labels[i],
                    )
        except:
            self.fail("Failed Execution: postprocess()")


if __name__ == "__main__":
    unittest.main()


"""
File containing a class for loading sampling data as a dataset, as well as a function for loading the dataset into a PyTorch dataloader.
"""
import torch
import math
import numpy as np
from torch.utils.data import Dataset, DataLoader, random_split
from brainspy.utils.pytorch import TorchUtils
from typing import Tuple, List


class ModelDataset(Dataset):
    def __init__(self, filename: str, steps: int = 1) -> None:
        """
        Initialisation of the dataset. It loads a posprocessed_data.npz file into memory.
        The targets of this file are divided by the amplification correction factor, so that
        data is made setup independent.

        Parameters
        ----------
        filename : str
            Folder and filename where the posprocessed_data.npz is.
        steps : int
            It allows to skip parts of the data when loading it into memory. The number indicates
            how many items will be skipped in between. By default, step number is one (no values
            are skipped). E.g., if steps = 2, and the inputs are [0, 1, 2, 3, 4, 5, 6]. The only
            inputs taken into account would be: [0, 2, 4, 6].

        Notes
        -----
        The postprocessed data is a .npz file called postprocessed_data.npz
        with keys: inputs, outputs and info (dict)

        1. inputs: np.array
        The input(s) is(are) gathered for all activation electrodes. The units is in Volts.

        2. outputs: The output(s) is(are) gathered from all the readout electrodes. The units are in nA.
        The output data is raw. Additional amplification correction might be needed, this is
        left for the user to decide.

        3. info: dict
        Data structure of output and input are arrays of NxD, where N is the number of samples
        and D is the dimension.

        The configs dictionary contains a copy of the configurations used for sampling the data.
        In addition, the configs dictionary has a key named electrode_info, which is created
        during the postprocessing step. The electrode_info key contains the following keys:
        3.1 electrode_no: int
        Total number of electrodes in the device

        3.2 activation_electrodes: dict

        3.2.1 electrode_no: int
        Number of activation electrodes used for gathering the data

        3.2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and
        maximum of the ranges, respectively.

        3.3 output_electrodes: dict
        
        3.3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        3.3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified
        values.

        3.3.3 amplification: float
        Amplification correction factor used in the device to correct the
        amplification applied to the output current in order to convert it into
        voltage before its readout.
        """
        self.inputs, targets, self.sampling_configs = self.load_data_from_npz(
            filename, steps)
        self.targets = (targets /
                        self.sampling_configs["driver"]["amplification"])
        self.inputs = TorchUtils.format(self.inputs)
        self.targets = TorchUtils.format(self.targets)

        assert len(self.inputs) == len(
            self.targets), "Inputs and Outpus have NOT the same length"

    def __len__(self) -> int:
        """
        Overwrittes the __len__ method from the super class torch.utils.data.

        Returns
        -------
        int
            Size of the whole dataset.
        """
        return len(self.inputs)

    def __getitem__(self, index: int) -> Tuple[np.array]:
        """
        Overwrittes the __getitem__ method from the super class torch.utils.data.
        The method supports fetching a data sample for a given key.

        Parameters
        ----------
        index : int
            Index corresponding to the place of the data in the dataset.

        Returns
        -------
        tuple
            Inputs and targets of the dataset corresponding to the given index.
        """
        return (self.inputs[index, :], self.targets[index, :])

    def load_data_from_npz(self, filename: str,
                           steps: int) -> Tuple[np.array, np.array, dict]:
        """
        Loads the inputs, targets and sampling configurations from a given postprocessed_data.npz
        file.

        Parameters
        ----------
        filename : str
            Folder and filename where the posprocessed_data.npz is.
        steps : int
            It allows to skip parts of the data when loading it into memory. The number indicates
            how many items will be skipped in between. By default, step number is one (no values
            are skipped). E.g., if steps = 2, and the inputs are [0, 1, 2, 3, 4, 5, 6]. The only
            inputs taken into account would be: [0, 2, 4, 6].

        Returns
        -------
        inputs : np.array
            Input waves sent to the activation electrodes of the device during sampling.
        outputs : np.array
            Raw output data from the readout electrodes of the device during sampling,
            corresponding to the input.
        sampling_configs : dict
            Dictionary containing the sampling configurations with which the data was
            acquired.

        Notes
        -----
        The postprocessed data is a .npz file called postprocessed_data.npz
        with keys: inputs, outputs and info (dict)

        1. inputs: np.array
        The input(s) is(are) gathered for all activation electrodes. The units is in Volts.

        2. outputs: The output(s) is(are) gathered from all the readout electrodes. The units are in nA.
        The output data is raw. Additional amplification correction might be needed, this is
        left for the user to decide.

        3. info: dict
        Data structure of output and input are arrays of NxD, where N is the number of samples
        and D is the dimension.

        The configs dictionary contains a copy of the configurations used for sampling the data.
        In addition, the configs dictionary has a key named electrode_info, which is created
        during the postprocessing step. The electrode_info key contains the following keys:
        3.1 electrode_no: int
        Total number of electrodes in the device

        3.2 activation_electrodes: dict

        3.2.1 electrode_no: int
        Number of activation electrodes used for gathering the data

        3.2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and
        maximum of the ranges, respectively.

        3.3 output_electrodes: dict
        
        3.3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        3.3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified
        values.

        3.3.3 amplification: float
        Amplification correction factor used in the device to correct the
        amplification applied to the output current in order to convert it into
        voltage before its readout.
        """
        print("\n* Loading data from file:\n" + filename)
        # Pickle = True, since it also contains a dictionary.
        with np.load(filename, allow_pickle=True) as data:
            sampling_configs = dict(data["sampling_configs"].tolist())
            inputs = data["inputs"][::steps]
            outputs = data["outputs"][::steps]
            print(
                f"\t- Shape of inputs:  {inputs.shape}\n\t- Shape of outputs: {outputs.shape}\n"
            )
            print(
                f"* Sampling configs has the following keys:\n\t{sampling_configs.keys()}\n"
            )
        return inputs, outputs, sampling_configs


def get_info_dict(training_configs: dict, sampling_configs: dict) -> dict:
    """
    Retrieve the info dictionary given the training configs and the sampling configs.
    Note that the electrode_info key should be present in the sampling configs. This
    key is automatically generated when postprocessing the data.

    Parameters
    ----------
    training_configs : dict
        A copy of the configurations used for training the surrogate model.
    sampling_configs : dict
        A copy of the configurations used for sampling the training data.

    Returns
    -------
    info_dict
        This dictionary is required in order to initialise a surrogate
        model. It contains the following keys:
        1. model_structure: dict
        The definition of the internal structure of the surrogate model, which is typically five
        fully-connected layers of 90 nodes each.

        1.1 hidden_sizes : list
        A list containing the number of nodes of each layer of the surrogate model.
        E.g., [90,90,90,90,90]
        
        1.2 D_in: int
        Number of input features of the surrogate model structure. It should correspond to
        the activation electrode number.

        1.3 D_out: int
        Number of output features of the surrogate model structure. It should correspond to
        the readout electrode number.

        2. electrode_info: dict
        It contains all the information required for the surrogate model about the electrodes.

        2.1 electrode_no: int
        Total number of electrodes in the device

        2.2 activation_electrodes: dict

        2.2.1 electrode_no: int
        Number of activation electrodes used for gathering the data

        2.2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and
        maximum of the ranges, respectively.

        2.3 output_electrodes: dict

        2.3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        2.3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified
        values.

        2.3.3 amplification: float
        Amplification correction factor used in the device to correct the
        amplification applied to the output current in order to convert it into
        voltage before its readout.

        3. training_configs: dict
        A copy of the configurations used for training the surrogate model.

        4. sampling_configs : dict
        A copy of the configurations used for gathering the training data.
    """
    info_dict = {}
    info_dict["model_structure"] = training_configs["model_structure"].copy()
    info_dict["electrode_info"] = sampling_configs["electrode_info"].copy()
    del training_configs["model_structure"]
    info_dict["training_configs"] = training_configs.copy()
    del sampling_configs["electrode_info"]
    info_dict["sampling_configs"] = sampling_configs.copy()
    return info_dict


def get_dataloaders(
        configs: dict
) -> Tuple[List[torch.utils.data.DataLoader], float, dict]:
    """
    Loads all the datasets specified in the dataset_paths list key of the configurations dictionary
    and creates a dataloader.

    Parameters
    ----------
    configs : dict
        Surrogate model generation configurations.

        1. results_base_dir: str
        Directory where the trained model and corresponding performance plots will be stored.

        2. seed: int
        Sets the seed for generating random numbers to a non-deterministic random number.

        3. hyperparameters:
        epochs: int
        learning_rate: float

        4. model_structure: dict
        The definition of the internal structure of the surrogate model, which is typically five
        fully-connected layers of 90 nodes each.

        4.1 hidden_sizes : list
        A list containing the number of nodes of each layer of the surrogate model.
        E.g., [90,90,90,90,90]

        4.2 D_in: int
        Number of input features of the surrogate model structure. It should correspond to
        the activation electrode number.

        4.3 D_out: int
        Number of output features of the surrogate model structure. It should correspond to
        the readout electrode number.

        5. data:
        5.1 dataset_paths: list[str]
        A list of paths to the Training, Validation and Test datasets, stored as
        postprocessed_data.npz. It also supports adding a single training dataset, and splitting
        it using the configuration split_percentages.

        5.2 split_percentages: list[float] (Optional)
        When provided together a single dataset path, in the dataset_paths list, this variable
        allows to split it into training, validation and test datasets by providing the split
        percentage values. E.g. [0.8, 0.2] will split the training dataset into 80% of the data
        for training and 20% of the data for validation. Similarly, [0.8, 0.1, 0.1] will split
        the training dataset into 80%, 10% for validation dataset and 10% for test dataset. Note
        that all split values in the list should add to 1.

        5.3 steps : int
        It allows to skip parts of the data when loading it into memory. The number indicates
        how many items will be skipped in between. By default, step number is one (no values
        are skipped). E.g., if steps = 2, and the inputs are [0, 1, 2, 3, 4, 5, 6]. The only
        inputs taken into account would be: [0, 2, 4, 6].

        5.4 batch_size: int
        How many samples will contain each forward pass.

        5.5 worker_no: int
        How many subprocesses to use for data loading. 0 means that the data will be loaded in
        the main process. (default: 0)

        5.6 pin_memory: boolean
        If True, the data loader will copy Tensors into CUDA pinned memory before returning
        them. If your data elements are a custom type, or your collate_fn returns a batch that
        is a custom type.

    Returns
    -------
    dataloaders : list[torch.utils.dataDataLoader]
        A list containing the corresponding training, validation and test datasets.

    """
    # Load dataset
    # Only training configs will be taken into account for info dict
    # For ranges and etc.
    assert 'data' in configs and 'dataset_paths' in configs['data']
    assert isinstance(configs['data']['dataset_paths'],
                      list), "Paths for datasets should be passed as a list"
    assert configs['data']['dataset_paths'] != [], "Empty paths for datasets"
    datasets = []
    info_dict = None
    amplification = None
    dataset_names = ['train', 'validation', 'test']
    if len(configs['data']['dataset_paths']) > 1:
        for i in range(len(configs['data']['dataset_paths'])):
            if configs['data']['dataset_paths'][i] is not None:
                dataset = ModelDataset(configs['data']['dataset_paths'][i],
                                       steps=configs['data']['steps'])

                if i > 0:
                    amplification_aux = TorchUtils.format(
                        info_dict["sampling_configs"]["driver"]
                        ["amplification"])
                    assert torch.eq(amplification_aux, amplification).all(), (
                        "Amplification correction factor should be the same for all datasets."
                        + "Check if all datasets come from the same setup.")
                    info_dict[dataset_names[i] +
                              '_sampling_configs'] = dataset.sampling_configs
                else:
                    info_dict = get_info_dict(configs,
                                              dataset.sampling_configs)
                amplification = TorchUtils.format(
                    info_dict["sampling_configs"]["driver"]["amplification"])
                datasets.append(dataset)
    else:
        dataset = ModelDataset(configs['data']['dataset_paths'][0],
                               steps=configs['data']['steps'])
        info_dict = get_info_dict(configs, dataset.sampling_configs)
        amplification = TorchUtils.format(
            info_dict["sampling_configs"]["driver"]["amplification"])

        assert sum(
            configs['data']
            ['split_percentages']) == 1, "Split percentages should add to one"
        assert len(configs['data']['split_percentages']) <= 3 and len(
            configs['data']['split_percentages']
        ) >= 1, "Split percentage list should only allow from one to three values. For training, validation datasets."
        if len(configs['data']['split_percentages']) == 1:
            datasets = [dataset, None]
        else:
            train_set_size = math.ceil(
                configs['data']['split_percentages'][0] * len(dataset))
            valid_set_size = math.floor(
                configs['data']['split_percentages'][1] * len(dataset))

            if len(configs['data']['split_percentages']) == 2:
                datasets = list(
                    random_split(dataset, [train_set_size, valid_set_size]))
            else:
                test_set_size = len(dataset) - train_set_size - valid_set_size
                datasets = list(
                    random_split(
                        dataset,
                        [train_set_size, valid_set_size, test_set_size]))

    # Create dataloaders
    dataloaders = []
    shuffle = [True, False, False]
    for i in range(len(datasets)):
        if datasets[i] is not None and len(datasets[i]) != 0:
            dl = DataLoader(
                dataset=datasets[i],
                batch_size=configs["data"]["batch_size"],
                num_workers=configs["data"]["worker_no"],
                pin_memory=configs["data"]["pin_memory"],
                shuffle=shuffle[i],
            )
            dl.tag = dataset_names[i]
            dataloaders.append(dl)
        else:
            dataloaders.append(None)
    return dataloaders, amplification, info_dict



import unittest
import numpy as np
from bspysmg.data import postprocess


class Test_PostProcess(unittest.TestCase):
    def test_get_sampling_data(self):
        with self.assertRaises(TypeError):
            inputs, outputs = postprocess.get_sampling_data()

        try:
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IO.dat", 7, 1)
        except Exception:
            self.fail("Failed Execution: get_sampling_data")

        with self.assertRaises(AssertionError):
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IO.dat", 3, 1)
        with self.assertRaises(OSError):
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IOtemp.dat", 1, 5)

        with self.assertRaises(AssertionError):
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IO.dat", 6.1, 1.9)
        with self.assertRaises(AssertionError):
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IO.dat", 3.1, 1.2)

        with self.assertRaises(AssertionError):
            inputs, outputs = postprocess.get_sampling_data(
                "tests/data/IO.dat", 6.1, 1.9)

    def test_type_error_clipping(self):
        data = np.load('tests/data/postprocessed_data.npz')
        with self.assertRaises(TypeError):
            postprocess.clip_data(data['inputs'],
                                  data['outputs'],
                                  clipping_value_range=34)

    def test_clipping_value_is_none(self):
        data = np.load('tests/data/postprocessed_data.npz')
        try:
            postprocess.clip_data(data['inputs'],
                                  data['outputs'],
                                  clipping_value_range=None)
        except Exception:
            self.fail("Failed execution with no clipping value")

    def test_post_process(self):
        with self.assertRaises(TypeError):
            inputs, output, configs = postprocess.post_process()

        try:
            inputs, output, configs = postprocess.post_process("tests/data/")
            inputs, output, configs = postprocess.post_process(
                "tests/data/", [-4, 4])
        except Exception:
            self.fail("Failed Execution: post_process()")

        with self.assertRaises(FileNotFoundError):
            inputs, outputs = postprocess.post_process("not_existing_file")

        with self.assertRaises(AssertionError):
            inputs, output, configs = postprocess.post_process(
                "tests/data/", [-4, 4], 41.5, 15.1)


if __name__ == "__main__":
    unittest.main()


"""
Module for running all tests on brainspy.
"""

import os
import unittest
import bspysmg

if __name__ == "__main__":
    from datetime import datetime
    import matplotlib
    matplotlib.use('Agg')

    timestamp = datetime.today().strftime('%d-%m-%Y-%H:%M:%S')

    bspysmg.TEST_MODE = 'SIMULATION_PC'

    modules_to_test = unittest.defaultTestLoader.discover(start_dir="tests/",
                                                          pattern="*.py",
                                                          top_level_dir=None)
    
    runner = unittest.TextTestRunner()
    runner.run(modules_to_test)
    print(bspysmg.TEST_MODE)



import unittest
from bspysmg.utils import inputs
import copy
import numpy as np


class Test_Inputs(unittest.TestCase):
    def __init__(self, *args, **kwargs) -> None:
        super(Test_Inputs, self).__init__(*args, **kwargs)
        self.configs = {}
        self.configs["driver"] = {
            "instrument_type": "cdaq_to_cdaq",
            "inverted_output": True,
            "instruments_setup": {
                "multiple_devices":
                False,
                "average_io_point_difference":
                True,
                "activation_sampling_frequency":
                1000,
                "readout_sampling_frequency":
                10000,
                "trigger_source":
                "cDAQ1/segment1",
                "activation_instrument":
                "cDAQ1Mod3",
                "activation_channels": [0, 2, 5, 3, 4, 6, 1],
                "activation_voltage_ranges": [[-0.55, 0.325], [-0.95, 0.55],
                                              [-1, 0.6], [-1, 0.6], [-1, 0.6],
                                              [-0.95, 0.55], [-0.55, 0.325]],
                "activation_channel_mask": [1, 1, 1, 1, 1, 1, 1],
                "readout_instrument":
                "cDAQ1Mod4",
                "readout_channels": [0]
            },
            "amplification": [39.5],
            "gain_info": "50MOhm"
        }
        self.configs["input_data"] = {
            "input_distribution": "sawtooth",
            "activation_electrode_no": 7,
            "readout_electrode_no": 1,
            "input_frequency": [2, 3, 5, 7, 13, 17, 19],
            "phase": [-0.57, 0.25, -1.54, 2.17, 0.08, 0.15, -0.65],
            "ramp_time": 0.5,
            "batch_time": 0.5,
            "number_batches": 1,
            "amplitude": [0.55, 0.95, 0.95, 0.95, 0.95, 0.95, 0.55],
            "offset": [-0.15, -0.25, -0.25, -0.25, -0.25, -0.25, -0.15]
        }

    def test_empty_inputs(self):
        with self.assertRaises(TypeError):
            configs, func = inputs.get_input_generator()

    def test_normal_inputs(self):
        CONFIGS = copy.deepcopy(self.configs)
        try:
            configs, func = inputs.get_input_generator(CONFIGS)
        except:
            self.fail("Exeution Failed")

    def test_configs_keys(self):

        with self.assertRaises(AssertionError):
            configs_4 = copy.deepcopy(self.configs)
            del configs_4["input_data"]['amplitude']
            configs, func = inputs.get_input_generator(configs_4)

        with self.assertRaises(AssertionError):
            configs_5 = copy.deepcopy(self.configs)
            del configs_5["input_data"]['offset']
            configs, func = inputs.get_input_generator(configs_5)

        try:
            configs_6 = copy.deepcopy(self.configs)
            configs_6['input_data']['input_distribution'] = 'sine'
            configs, func = inputs.get_input_generator(configs_6)
        except Exception:
            self.fail('Failed initialising sinewave')

        with self.assertRaises(NotImplementedError):
            configs_7 = copy.deepcopy(self.configs)
            configs_7['input_data']['input_distribution'] = 'asdf'
            configs, func = inputs.get_input_generator(configs_7)

    def test_sine_wave(self):
        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave()

        try:
            sine_wave = inputs.sine_wave(np.array(range(0)), 0.1, 5, 0.1, 10)
            sine_wave = inputs.sine_wave(np.array(range(100)), 60, 0, 5, 0)
            sine_wave = inputs.sine_wave(np.array(range(100)), 0.1, 0, 5, 0)
            sine_wave = inputs.sine_wave(np.array(range(100)), 0.1, 5, 5, 0)
            sine_wave = inputs.sine_wave(np.array(range(100)), 0.1, 5, 0.1, 0)
            sine_wave = inputs.sine_wave(np.array(range(100)), 0.1, 5, 0.1, 10)
            assert sine_wave.flatten().shape[0] == 100
        except:
            self.fail("Failed creating sine_wave")

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(list(range(100)), 60, 0, 5, 0)

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(100, 60, 0, 5, 0)

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(np.array(range(100)), "60", 0, 5, 0)

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(np.array(range(100)), 60, "5", 5, 0)

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(np.array(range(100)), 60, 0, "5", 0)

        with self.assertRaises(TypeError):
            sine_wave = inputs.sine_wave(np.array(range(100)), 60, 0, 5, "10")

    def test_sawtooth_wave(self):
        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave()

        try:
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 60, 0,
                                                 5, 0)
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 0.1, 0,
                                                 5, 0)
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 0.1, 5,
                                                 5, 0)
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 0.1, 5,
                                                 0.1, 0)
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 0.1, 5,
                                                 0.1, 10)
            assert sawtooth_wave.flatten().shape[0] == 100
        except:
            self.fail("Failed creating sawtooth_wave")

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(list(range(100)), 60, 0, 5, 0)

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(100, 60, 0, 5, 0)

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), "60", 0,
                                                 5, 0)

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 60, "5",
                                                 5, 0)

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 60, 0,
                                                 "5", 0)

        with self.assertRaises(TypeError):
            sawtooth_wave = inputs.sawtooth_wave(np.array(range(100)), 60, 0,
                                                 5, "10")

    def test_generate_sinewave(self):
        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave()

        try:
            sinewave = inputs.generate_sinewave(100, 60, 5, 0)
            sinewave = inputs.generate_sinewave(20, 0.1, 5, 0)
            sinewave = inputs.generate_sinewave(100, 0.1, 5, 5)
            sinewave = inputs.generate_sinewave(100, 0.1, 5, 0.1)
            assert sinewave.flatten().shape[0] == 100
        except:
            self.fail("Failed creating generate_sinewave")

        with self.assertRaises(ZeroDivisionError):
            sinewave = inputs.generate_sinewave(0, 60, 0, 5)

        with self.assertRaises(ValueError):
            sinewave = inputs.generate_sinewave(-10, 60, 0, 5)

        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave(40.0, 60, 0, 5)

        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave(np.array(range(100)), 60, 0, 5)

        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave(100, "60", 0, 5)

        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave(100, 60, "5", 5)

        with self.assertRaises(TypeError):
            sinewave = inputs.generate_sinewave(100, 60, 0, "5")

    def test_generate_sawtooth_simple(self):
        with self.assertRaises(TypeError):
            sawtooth_simple = inputs.generate_sawtooth_simple()

        try:
            sawtooth_simple = inputs.generate_sawtooth_simple(
                -5, 5, 100, False)
            sawtooth_simple = inputs.generate_sawtooth_simple(-5, 5, 100, True)
            sawtooth_simple = inputs.generate_sawtooth_simple(-6.5, 7.5, 10)
            assert sawtooth_simple.flatten().shape[0] == 10
        except:
            self.fail("Failed creating generate_sinewave")

        with self.assertRaises(ValueError):
            sawtooth_simple = inputs.generate_sawtooth_simple(5, 15, 20)

        with self.assertRaises(AssertionError):
            sawtooth_simple = inputs.generate_sawtooth_simple(5, -5, 20)

        with self.assertRaises(AssertionError):
            sawtooth_simple = inputs.generate_sawtooth_simple(-5, 5, 5)

    def test_generate_sawtooth_multiple(self):
        with self.assertRaises(TypeError):
            sawtooth_simple = inputs.generate_sawtooth_multiple()

        try:
            sawtooth_simple = inputs.generate_sawtooth_multiple([-5, 5], 100,
                                                                False)
            # sawtooth_simple = inputs.generate_sawtooth_multiple([-5, 5], 100,
            #                                                     True)
            sawtooth_simple = inputs.generate_sawtooth_multiple([-6.5, 7.5],
                                                                10)
            # assert sawtooth_simple.flatten().shape[0] == 10
        except Exception:
            self.fail("Failed creating generate_sinewave")

    def test_generate_random_phase(self):
        phase = inputs.get_random_phase()
        self.assertEquals(len(phase), 7)
        phase2 = inputs.get_random_phase()
        self.assertTrue((phase != phase2).all())


if __name__ == '__main__':
    unittest.main()


"""
File containing functions for postprocessing raw data gathered from the sampler and information for the model's info dictionary.
"""
import os

import numpy as np

from brainspy.utils.io import load_configs
from bspysmg.utils.plots import output_hist
from typing import Tuple
from datetime import timedelta


def get_sampling_data(filename: str, activation_electrode_no: int,
                      readout_electrode_no: int) -> Tuple[np.array, np.array]:
    """
    Reads the sampling data from a text file (IO.dat) and returs the values loaded in numpy arrays.

    Parameters
    ----------
    filename : str
        Path to the file containing comma separated values read during the data gathering
        process. Typically, named IO.dat.
    activation_electrode_no : int
        Number of activation electrodes used for the device during the data gathering process.
    readout_electrode_no : int
        Number of current readout/output electrodes used for the device during the data gathering
        process.

    Returns
    -------
    inputs : np.array
        Array containing all the inputs that were sent to the device during sampling.
    outputs : np.array
        Array containing all the outputs of the device obtained during sampling, which correspond
        to the inputs to the device.
    """
    print("\nLoading file: " + filename)
    print("This may take some time. Please wait.\n")
    assert type(activation_electrode_no
                ) is int, "Activation electrode number expected to be int"
    assert type(readout_electrode_no
                ) is int, "Readout electrode number expected to be int"
    data = np.loadtxt(filename)
    assert data.shape[1] == (
        activation_electrode_no + readout_electrode_no
    ), "Data from the file has a different electrode configuration. Check the activation electrode no and the readout electrode no"
    inputs = data[:, :activation_electrode_no]
    outputs = data[:, -readout_electrode_no:]
    return inputs, outputs


def post_process(data_dir: str,
                 clipping_value="default",
                 charging_signal_batch_no: int = 40,
                 reference_signal_batch_no: int = 15,
                 filename: str = "postprocessed_data",
                 **kwargs) -> Tuple[np.array, np.array, dict]:
    """
    Postprocesses the data, cleans any clipping (optional), and merges data sets if needed. The data
    arrays are merged into a single array and cropped given the clipping_values. The function also
    plots and saves the histogram of the data.

    Parameters
    ----------
    data_dir: str
        A string with path to the directory with the data: it is assumed at least two
        files exist, named sampler_configs.json and a IO.dat respectively.
    clipping_value : [float,float]
        Will apply a clipping to the input and output sampling data within the
        specified values. The the setups have a limit in the range they can read.
        They typically clip at approximately +-4 V. Note that in order to
        calculate the clipping_range, it needs to be multiplied by the
        amplification value of the setup. (e.g., in the Brains setup the
        amplification is 28.5, is the clipping_value is +-4 (V), therefore, the
        clipping value should be +-4 * 28.5, which is [-110,110] (nA) ).
        This variable represents a lower and upper clipping_value to crop data.
        It can be either None, 'default' or [float,float]. The 'default' str
        input will automatically take the clipping value by multiplying the
        amplification of the data by -4 and 4. The None input will not apply any
        clipping. 
        
        N O T E: When the clipping value is set to None, the model will accurately
        represent the hardware setup (feedback resistance of the operational
        amplifier). When clipping value set to the values that
        are clipping, the model will extrapolate the results outside of the clipping
        range caused by the hardaware setup.
    charging_signal_batch_no: [int]
        Number of batches that will be used for extracting the charging signal.
    reference_signal_batch_no: [int]
        Number of batches that will be used for extracting the reference signal.
    filename: [str]
        The name of the file that will be produced after postprocessing. By default: postprocessed_data.npz
    kwargs: Optional kwargs are as follows:
        1. list_data: A list of strings indicating directories with postprocessed_data.npz
        containing input and output data relationships from the device, as well
        as the configuration with which the data was acquired.

    Examples
    --------

    >>> inputs, outputs, configs = post_process('tmp/data/training/TEST/17-02-2021/')

    Notes
    -----
    The postprocessed data is a .npz file called postprocessed_data.npz
    with keys: inputs, outputs and info (dict)

    1. inputs: np.array
    The input(s) is(are) gathered for all activation electrodes. The units is in Volts.

    2. outputs: The output(s) is(are) gathered from all the readout electrodes. The units are in nA.
    The output data is raw. Additional amplification correction might be needed, this is
    left for the user to decide.

    3. info: dict
    Data structure of output and input are arrays of NxD, where N is the number of samples
    and D is the dimension.

    The configs dictionary contains a copy of the configurations used for sampling the data.
    In addition, the configs dictionary has a key named electrode_info, which is created
    during the postprocessing step. The electrode_info key contains the following keys:
    3.1 electrode_no: int
    Total number of electrodes in the device

    3.2 activation_electrodes: dict

    3.2.1 electrode_no: int
    Number of activation electrodes used for gathering the data

    3.2.2 voltage_ranges: list
    Voltage ranges used for gathering the data. It contains the ranges per
    electrode, where the shape is (electrode_no,2). Being 2 the minimum and
    maximum of the ranges, respectively.

    3.3 output_electrodes: dict
    
    3.3.1 electrode_no : int
    Number of output electrodes used for gathering the data

    3.3.2 clipping_value: list[float,float]
    Value used to apply a clipping to the sampling data within the specified
    values.

    3.3.3 amplification: float
    Amplification correction factor used in the device to correct the
    amplification applied to the output current in order to convert it into
    voltage before its readout.

    """
    assert type(charging_signal_batch_no
                ) is int, "charging_signal_batch_no should be an integer"
    assert type(reference_signal_batch_no
                ) is int, "reference_signal_batch_no should be an integer"
    assert data_dir is not None
    configs = load_configs(os.path.join(data_dir, "sampler_configs.json"))
    activation_electrode_no = configs["input_data"]["activation_electrode_no"]
    readout_electrode_no = configs["input_data"]["readout_electrode_no"]

    # If the data comes from multiple sources. Merge them first.
    # if "list_data" in kwargs.keys():
    #     inputs, outputs, configs = data_merger(
    #         data_dir,
    #         kwargs["list_data"],
    #         activation_electrode_no=activation_electrode_no,
    #         readout_electrode_no=readout_electrode_no)
    # elif len(kwargs.keys()) > 0:
    #     assert (
    #         False
    #     ), f"{list(kwargs.keys())} not recognized! kwargs must be list_data"

    inputs, outputs = get_sampling_data(
        os.path.join(data_dir, "IO.dat"),
        activation_electrode_no=activation_electrode_no,
        readout_electrode_no=readout_electrode_no)

    batch_length = int(
        configs["input_data"]["batch_time"] *
        configs["driver"]["instruments_setup"]["activation_sampling_frequency"]
    )
    nr_raw_samples = len(outputs)
    print("Number of raw samples: ", nr_raw_samples)
    assert (nr_raw_samples == configs["input_data"]["number_batches"] *
            batch_length), "Data size mismatch!"
    output_scales = [np.min(outputs), np.max(outputs)]
    print(f"Output scales: [Min., Max.] = {output_scales}")
    # input_scales = list(zip(np.min(inputs, axis=0), np.max(inputs, axis=0)))
    print(f"Lower bound input scales: {np.min(inputs,axis=0)}")
    print(f"Upper bound input scales: {np.max(inputs,axis=0)}\n")
    # Get charging signals
    # charging_batches = int(
    #     60 * 30 /
    #     configs["input_data"]["batch_time"])  # ca. 30 min charging signal
    print("Charging signal contains " + str(charging_signal_batch_no) +
          " batches. Total time: " + str(
              timedelta(seconds=int(charging_signal_batch_no *
                                    configs['input_data']['batch_time']))))
    save_npz(
        data_dir,
        "charging_signal",
        inputs[-charging_signal_batch_no * batch_length:],
        outputs[-charging_signal_batch_no * batch_length:],
        configs,
    )
    # # Get reference batches
    # refs_batches = int(
    #     600 / configs["input_data"]["batch_time"])  # ca. 600s reference signal
    print("\nReference signal contains " + str(reference_signal_batch_no) +
          " batches. Total time: " + str(
              timedelta(seconds=int(reference_signal_batch_no *
                                    configs['input_data']['batch_time']))))
    save_npz(
        data_dir,
        "reference_batch",
        inputs[-reference_signal_batch_no * batch_length:],
        outputs[-reference_signal_batch_no * batch_length:],
        configs,
    )
    # Plot samples histogram and save
    output_hist(outputs[::3], data_dir, bins=100)

    # Clean data
    configs["electrode_info"] = get_electrode_info(configs, clipping_value)
    if configs["electrode_info"]["output_electrodes"][
            "clipping_value"] is not None:
        inputs, outputs = clip_data(
            inputs,
            outputs,
            configs["electrode_info"]["output_electrodes"]["clipping_value"],
        )
        print("% of points cropped: ",
              (1 - len(outputs) / nr_raw_samples) * 100)
        print("\n")
    # save data
    save_npz(data_dir, filename, inputs, outputs, configs)

    return inputs, outputs, configs


def save_npz(data_dir: str, file_name: str, inputs: np.array,
             outputs: np.array, configs: dict) -> None:
    """
    Stores the input, outputs and sampling configurations in an .npz file.
    The saved file needs to be opened with the option pickle=True, since it
    contains a dictionary.

    Parameters
    ----------
    data_dir : str
        Folder where the data is going to be stored.
    file_name : [type]
        The name of the data that wants to be stored.
    inputs : np.array
        Array containing all the inputs that were sent to the device during sampling.
    outputs : np.array
        Array containing all the outputs of the device obtained during sampling, which correspond
        to the inputs to the device.
    configs : dict
        Sampling configurations with the following keys:

        1. save_directory: str
        Directory where the all the sampling data will be stored.

        2. data_name: str
        Inside the path specified on the variable save_directory, a folder will be created,
        with the format: <data_name>+<current_timestamp>. This variable specified the
        prefix of that folder before the timestamp.

        3. driver: dict
        Dictionary containing the driver configurations. For more information check the
        documentation about this configuration file, check the documentation of
        brainspy.processors.hardware.drivers.ni.setup.NationalInstrumentsSetup

        4. input_data : dict
        Dictionary containing the information necessary to create the input sampling data.
        4.1 input_distribution: str
        It determines the wave shape of the input. Two main options availeble 'sawtooth'
        and 'sine'. The first option will create saw-like signals, and the second
        sine-wave signals. Sawtooth signals have more coverage on the edges of the
        input range.

        4.2 activation_electrode_no: int
        Number of activation electrodes in the device that wants to be sampled.

        4.3 readout_electrode_no : int
        Number of readout electrodes in the device that wants to be sampled.

        4.4 input_frequency: list
        Base frequencies of the input waves that will be created. In order to optimise
        coverage, irrational numbers are recommended. The list should have the same
        length as the activation electrode number. E.g., for 7 activation electrodes:
        input_frequency = [2, 3, 5, 7, 13, 17, 19]

        4.5 phase : float
        Horizontal shift of the input signals. It is recommended to have random numbers
        which are different for the training, validation and test datasets. These
        numbers will be square rooted and multiplied by a given factor.

        4.6 factor : float
        Given factor by which the input frequencies will be multiplied after square
        rooting them.

        4.7 amplitude : Optional[list[float]]
        Amplitude of the generated input wave signal. It is calculated according to the
        minimum and maximum ranges of each electrode. Where the amplitude value should
        correspond with (max_range_value - min_range_value) / 2. If no amplitude is
        given it will be automatically calculated from the driver configurations for
        activation electrode ranges. If it wants to be manually set, the offset
        variable should also be included in the dictionary.

        4.8 offset: Optional[list[float]]
        Vertical offset of the generated input wave signal. It is calculated according
        to the minimum and maximum ranges of each electrode. Where the offset value
        should correspond with (max_range_value + min_range_value) / 2. If no offset
        is given it will be automatically calculated from the driver configurations for
        activation electrode ranges. If it wants to be manually set, the offset
        variable should also be included in the dictionary.

        4.9 ramp_time: float
        Time that will be taken before sending each batch to go from zero to the first
        point of the batch and to zero from the last point of the batch.

        4.10 batch_time:
        Time that the sampling of each batch will take.

        4.11 number_batches: int
        Number of batches that will be sampled. A default value of 3880 is reccommended.
    """
    save_to = os.path.join(data_dir, file_name)
    print(f"Data saved to: {save_to}.npz")
    np.savez(save_to, inputs=inputs, outputs=outputs, sampling_configs=configs)


def get_electrode_info(configs: dict, clipping_value) -> dict:
    """
    Retrieve electrode information from the data sampling configurations.

    Parameters
    ----------
    configs : dict
        Sampling configurations with the following keys:
        1. driver: dict
        Dictionary containing the driver configurations. For more information check the
        documentation about this configuration file, check the documentation of
        brainspy.processors.hardware.drivers.ni.setup.NationalInstrumentsSetup
        
        2. input_data : dict
        Dictionary containing the information necessary to create the input sampling data.
        2.1 activation_electrode_no: int
        Number of activation electrodes in the device that wants to be sampled.

        2.2 readout_electrode_no : int
        Number of readout electrodes in the device that wants to be sampled.

        2.3 amplitude : [list[float]]
        Amplitude of the generated input wave signal. It is calculated according to the
        minimum and maximum ranges of each electrode. Where the amplitude value should
        correspond with (max_range_value - min_range_value) / 2. If no amplitude is
        given it will be automatically calculated from the driver configurations for
        activation electrode ranges. If it wants to be manually set, the offset
        variable should also be included in the dictionary.

        2.4 offset: [list[float]]
        Vertical offset of the generated input wave signal. It is calculated according
        to the minimum and maximum ranges of each electrode. Where the offset value
        should correspond with (max_range_value + min_range_value) / 2. If no offset
        is given it will be automatically calculated from the driver configurations for
        activation electrode ranges. If it wants to be manually set, the offset
        variable should also be included in the dictionary.
    clipping_value : str or list
        The value that will be used to clip the sampling data within a specific range. if
        default is passed, a default clipping value will be used. 

    Returns
    -------
    electrode_info : dict
        Configuration dictionary containing all the keys related to the electrode information:
        1. electrode_no: int
        Total number of electrodes in the device

        2. activation_electrodes: dict
        2.1 electrode_no: int
        Number of activation electrodes used for gathering the data

        2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and maximum
        of the ranges, respectively.

        3. output_electrodes: dict
        3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified values.

        3.3 amplification: float
        Amplification correction factor used in the device to correct the amplification
        applied to the output current in order to convert it into voltage before its
        readout.
    """
    electrode_info = {}
    electrode_info["electrode_no"] = (
        configs["input_data"]["activation_electrode_no"] +
        configs["input_data"]["readout_electrode_no"])
    electrode_info["activation_electrodes"] = {}
    electrode_info["activation_electrodes"]["electrode_no"] = configs[
        "input_data"]["activation_electrode_no"]
    electrode_info["activation_electrodes"][
        "voltage_ranges"] = get_voltage_ranges(
            configs["input_data"]["offset"],
            configs["input_data"]["amplitude"])
    electrode_info["output_electrodes"] = {}
    electrode_info["output_electrodes"]["electrode_no"] = configs[
        "input_data"]["readout_electrode_no"]
    electrode_info["output_electrodes"]["amplification"] = configs["driver"][
        "amplification"]
    if clipping_value == "default":
        electrode_info["output_electrodes"]["clipping_value"] = (
            electrode_info["output_electrodes"]["amplification"] *
            np.array([-4, 4])).tolist()
    else:
        electrode_info["output_electrodes"]["clipping_value"] = clipping_value

    print_electrode_info(electrode_info)
    return electrode_info


def get_voltage_ranges(offset: list, amplitude: list) -> np.array:
    """
    Calculate the voltage ranges of the device out of the information about the
    amplitude and the vertical offset that was used to compute the input waves
    during the data gathering process.

    Parameters
    ----------
    offset : list
        A list of all the offset values to vertically displace the input signal
        in such a way that it fits the activation electrode ranges. The list would
        contain one value per activation electrode.
    amplitude : list
        A list of all the amplitude values to amplify the input signal in such
        a way that it fits the activation electrode ranges.

    Returns
    -------
    np.array
        Array containing the ranges per electrode, where the shape is (electrode_no,2). Being
        2 the minimum and maximum of the ranges, respectively.
    """
    offset = np.array(offset, dtype=np.float32)
    amplitude = np.array(amplitude, dtype=np.float32)
    min_voltage = (offset - amplitude)[:, np.newaxis]
    max_voltage = (offset + amplitude)[:, np.newaxis]
    return np.concatenate((min_voltage, max_voltage), axis=1)


def print_electrode_info(configs: dict) -> None:
    """
    Prints on screen the information about the electrodes that was gathered
    from the configuration file used for gathering the data from the device.

    Parameters
    ----------
    configs : dict
        Configuration dictionary containing all the keys related to the electrode information:
        1. electrode_no: int
        Total number of electrodes in the device

        2. activation_electrodes: dict
        2.1 electrode_no: int
        Number of activation electrodes used for gathering the data

        2.2 voltage_ranges: list
        Voltage ranges used for gathering the data. It contains the ranges per
        electrode, where the shape is (electrode_no,2). Being 2 the minimum and maximum
        of the ranges, respectively.
        3. output_electrodes: dict
        3.1 electrode_no : int
        Number of output electrodes used for gathering the data

        3.2 clipping_value: list[float,float]
        Value used to apply a clipping to the sampling data within the specified values.

        3.3 amplification: float
        Amplification correction factor used in the device to correct the amplification
        applied to the output current in order to convert it into voltage before its
        readout.
    """
    print(
        "\nThe following data is inferred from the input data. Please check if it is correct. "
    )
    print(
        f"Data is gathered from a device with {configs['electrode_no']} electrodes, from which: "
    )
    print(
        f"There are {configs['activation_electrodes']['electrode_no']} activation electrodes: "
    )
    print("\t * Lower bound of voltage ranges: " +
          str(configs["activation_electrodes"]["voltage_ranges"][:, 0]))
    print("\t * Upper bound of voltage ranges: " +
          str(configs["activation_electrodes"]["voltage_ranges"][:, 1]))
    print(
        f"There are {configs['output_electrodes']['electrode_no']} output electrodes: "
    )
    print("\t * Clipping value: " +
          str(configs["output_electrodes"]["clipping_value"]))
    print("\t * Amplification correction value: " +
          str(configs["output_electrodes"]["amplification"]))


def clip_data(inputs: np.array, outputs: np.array,
              clipping_value_range: list) -> Tuple[np.array, np.array]:
    """
    Removes all the outputs and corresponding inputs where the output is outside a given maximum
    and minimum range.

    Parameters
    ----------
    inputs : np.array
        Array containing all the inputs that were sent to the device during sampling.
    outputs : np.array
        Array containing all the outputs of the device obtained during sampling, which correspond
        to the inputs to the device.
    clipping_value_range : list[float,float]
        A list of length two. The first element will be the lower clipping range, and the second
        element will be the higher clipping range.

    Returns
    -------
    inputs : np.array
        Array containing all the inputs that were sent to the device during sampling, except for
        those  values for which its corresponding output is above and below the specified clipping
        range.
    outputs : np.array
        Array containing all the outputs of the device obtained during sampling, except for those
        values for which its corresponding output is above and below the specified clipping range.
    """

    mean_output = np.mean(outputs, axis=1)

    # Get cropping mask
    if type(clipping_value_range) is list:
        cropping_mask = (mean_output < clipping_value_range[1]) * (
            mean_output > clipping_value_range[0])
        print(
            f"\nClipping data outside range {clipping_value_range[0]} and {clipping_value_range[1]}"
        )
        outputs = outputs[cropping_mask]
        inputs = inputs[cropping_mask, :]
        return inputs, outputs
    elif clipping_value_range is None:
        return inputs, outputs
    else:
        raise TypeError(
            f"Clipping value not recognized! Must be list with lower and upper bound or float, was {type(clipping_value_range)}"
        )


# def merge_postprocessed_data(file_names,
#                              output_file_name='merged_postprocessed_data.npz'):
#     """[summary]

#     Parameters
#     ----------
#     file_names : [type]
#         [description]
#     output_file_name : str, optional
#         [description], by default 'merged_postprocessed_data.npz'

#     Example
#     ----------
#     file_names = ['tmp/data/training/Brains_testing_2020_09_04_182557/postprocessed_data.npz',
#      'tmp/data/training/Brains_testing_2020_09_11_093200/postprocessed_data.npz']
#     merge_postprocessed_data(file_names)
#     """
#     ref_data = dict(np.load(file_names[0], allow_pickle='True'))
#     for i in range(1, len(file_names)):
#         data = np.load(file_names[i])
#         for key in list(data):
#             if key != 'info':
#                 ref_data[key] = np.append(ref_data[key], data[key], axis=0)
#     np.savez(output_file_name, **ref_data)

# def data_merger(main_dir, activation_electrode_no=7, readout_electrode_no=1):
#     # EXAMPLE
#     #  main_dir = "tmp/output/model_nips"
#     # The post_process function should have a clipping value which is in an amplified scale.
#     # E.g., for an amplitude of 100 -> 345.5
#     # process_multiple(main_dir)
#     shape = 0
#     dirs = list([
#         name for name in os.listdir(main_dir)
#         if os.path.isdir(os.path.join(main_dir, name))
#         and not name.startswith('.')
#     ])

#     assert len(dirs) > 0
#     for i in range(len(dirs)):
#         shape += np.load(os.path.join(main_dir, dirs[i],
#                                       'postprocessed_data.npz'),
#                          allow_pickle=True)['inputs'].shape[0]

#     input_results = np.zeros([shape, activation_electrode_no])
#     output_results = np.zeros([shape, readout_electrode_no])
#     previous_shape = 0
#     for i in range(len(dirs)):
#         data = np.load(os.path.join(main_dir, dirs[i],
#                                     'postprocessed_data.npz'),
#                        allow_pickle=True)
#         current_shape = previous_shape + data['inputs'].shape[0]
#         input_results[previous_shape:current_shape] = data['inputs']
#         output_results[previous_shape:current_shape] = data['outputs']
#         previous_shape = current_shape
#         info = data['info']

#     info = dict(np.ndenumerate(info))[()]
#     info['input_data']['input_distribution'] = 'mixed'
#     info['input_data']['phase'] = 'mixed'
#     index = np.random.permutation(np.arange(shape))
#     input_results = input_results[index]
#     output_results = output_results[index]

#     limit = int(shape * 0.75)

#     np.savez(os.path.join(main_dir, 'training_data'),
#              inputs=input_results[:limit],
#              outputs=output_results[:limit],
#              info=info)
#     np.savez(os.path.join(main_dir, 'test_data'),
#              inputs=input_results[limit:],
#              outputs=output_results[limit:],
#              info=info)

#if __name__ == "__main__":
# import matplotlib

# matplotlib.use('TkAgg')
# main_dir = "C:/Users/Unai/Documents/github/brainspy-smg/tmp/brains_setup/sampling_data_1KSPS_arsenic_test_2022_03_17_171248"
# inputs, outputs, info = post_process(main_dir,
#                                      clipping_value=[-114, 114],
#                                      filename="postprocessed_data_clipped")
# dirs = list(
#     [
#         name
#         for name in os.listdir(main_dir)
#         if os.path.isdir(os.path.join(main_dir, name)) and not name.startswith(".")
#     ]
# )

# assert len(dirs) > 0
# for i in range(len(dirs)):
# inputs, outputs, info = post_process(main_dir)
# output_hist(outputs, os.path.join(main_dir, dirs[i]), bins=1000, show=True)



"""
File containing several methods for plotting purposes.
"""
import os
import numpy as np
import matplotlib.pyplot as plt

import torch
from brainspy.utils.waveform import WaveformManager


def plot_error_hist(targets: np.array,
                    prediction: np.array,
                    error: np.array,
                    mse: np.array,
                    save_dir: str,
                    name: str = "error") -> None:
    """
    Plots and saves error histogram graph for between given target and predicted.

    Parameters
    ----------
    targets : np.array
        Reference data used for training/validation.
    prediction : np.array
        Predictions made by model.
    error : np.array
        Errors correspoding to each target data point.
    mse : np.array
        Mean squared error correspoding to each target data point.
    save_dir : str
        Name of the path where the graph is to be saved.
    name : str [Optional]
        Name of the file for the graph.
    """
    assert targets.size == prediction.size
    assert targets.size == error.size
    assert mse >= 0
    plt.figure()
    plt.title('Predicted vs True values')
    plt.subplot(1, 2, 1)
    plt.plot(targets, prediction, ".")
    plt.xlabel("True Output (nA)")
    plt.ylabel("Predicted Output (nA)")
    targets_and_prediction_array = np.concatenate((targets, prediction))
    min_out = np.min(targets_and_prediction_array)
    max_out = np.max(targets_and_prediction_array)
    plt.plot(np.linspace(min_out, max_out), np.linspace(min_out, max_out), "k")
    plt.title(f"RMSE {np.sqrt(mse)} (nA)")
    plt.subplot(1, 2, 2)
    plt.hist(np.reshape(error, error.size), 500)
    x_lim = 0.25 * np.max([np.abs(error.min()), error.max()])
    plt.xlim([-x_lim, x_lim])
    plt.title("Error histogram (nA) ")
    fig_loc = os.path.join(save_dir, name)
    plt.tight_layout()
    plt.savefig(fig_loc, dpi=300)
    plt.close()


def plot_error_vs_output(targets: np.array,
                         error: np.array,
                         save_dir: str,
                         name: str = "error_vs_output") -> None:
    """
    Plots and saves error vs output graph for given error and their
    correspoding output.

    Parameters
    ----------
    targets : np.array
        Reference data used for training/validation.
    error : np.array
        Errors correspoding to each target data point.
    save_dir : str
        Name of the path where the graph is to be saved.
    name : str [Optional]
        Name of the file for the graph.
    """
    assert targets.size == error.size
    plt.figure()
    plt.plot(targets, error, ".")
    plt.plot(
        np.linspace(
            targets.min(),
            targets.max(),
            len(error),
        ),
        np.zeros_like(error),
    )
    plt.title("Error vs Output")
    plt.xlabel("Output (nA)")
    plt.ylabel("Error (nA)")
    fig_loc = os.path.join(save_dir, name)
    plt.savefig(fig_loc, dpi=300)
    plt.close()


def plot_waves(inputs: np.array, outputs: np.array, input_no: int,
               output_no: int, batch: int, legend: np.array,
               save_directory: str) -> None:
    """
    Plots and saves input and output waves for the model. Image is overwritten 
    for each batch and it is used to control what is happening to the device
    input output relationship for each batch.

    Parameters
    ----------
    inputs : np.array
        Input data used for training/validation of model.
    outputs : np.array
        Output generated by model corresponding the inputs.
    input_no : int
        Input activation electrode number.
    output_no : int
        Output electrode number.
    batch : int
        The current batch number of the data.
    legend : np.array
        List of headers of text file which contains data used
        for training, validation, and testing.
    save_directory : str
        Name of the file for the graph.
    """
    plt.figure()
    plt.suptitle(f'I/O data for batch {batch}')
    plt.subplot(211)
    plt.plot(inputs)
    plt.ylabel('Inputs (V)')
    plt.xlabel('Time points (a.u.)')
    plt.legend(legend[:input_no])
    plt.subplot(212)
    plt.plot(outputs)
    plt.ylabel('Outputs (nA)')
    plt.legend(legend[-output_no:])
    plt.xlabel('Time points (a.u.)')
    plt.tight_layout()
    plt.savefig(os.path.join(save_directory, 'example_batch'))
    plt.close()


def output_hist(outputs: np.array,
                data_dir: str,
                bins: int = 100,
                show: bool = False) -> None:
    """
    Saves and optionally plots the histogram of output/predictions
    of the model.

    Parameters
    ----------
    outputs : np.array
        Output generated by model.
    data_dir : str
        Name of the path where the graph is to be saved.
    bins : int [Optional]
        Number of bins for the histogram.
    show : bool [Optional]
        If set to true, it displays the generated histogram.
    """
    plt.figure()
    plt.title("Output Histogram")
    plt.hist(outputs, bins=bins)
    plt.ylabel("Counts")
    plt.xlabel("Raw output (nA)")
    if show:
        plt.show()
    plt.savefig(data_dir + "/output_distribution")
    plt.close()


# def plot(self, x, y):
#     for i in range(np.shape(y)[1]):
#         plt.figure()
#         plt.plot(x)
#         plt.plot(y)
#         plt.show()


def iv_plot(result: np.array,
            inputs: np.array,
            input_electrode: int,
            data_dir: str = ".",
            save_plot: bool = None,
            show_plot: bool = False) -> None:
    """
    Plots IV characteristics and optinally saves the graph
    for a given electrode number.

    Parameters
    ----------
    result : np.array
        Output current values of an electrode.
    input_electrode : int
        Electrode number.
    save_plot : bool [Optional]
        If set to true, it saves the generated plot to
        current directory.
    show_plot : bool [Optional]
        If set to true, it displays the generated plot.
    """
    plt.plot(inputs,
             result,
             label='IV Curve for electrode ' + str(input_electrode))
    plt.xlabel('Voltage (V)')
    plt.ylabel('Current (nA)')
    if save_plot is not None:
        plt.savefig(data_dir + "/iv_plot")
    if show_plot:
        plt.show()


def multi_iv_plot(configs, inputs, output, save_plot=None, show_plot=True):
    """
    Plots the IV curve of several devices in one plot. Devices can be the DNPU
    device or a surrogate model.

    Parameters
    ----------
    configs : dict
        Dictionary containing the configurations for IV measurements with
        the following keys:
        
        1. devices: list
        List of devices for which IV response is to be computed. This list
        contains the names of all the devices (A,B,C,D etc) involved in the
        experiment.
        
        2. driver: dict
        It contains the configurations for each device in the experiment which
        are defined in the devices list.
    inputs : dict
        Dictionary containing the list of input signal waves for each device.
    outputs : dict
        Dictionary containing the list of output currents for each device.
    save_plot : str or None
        If None, the plot will not be saved, if a string provided, the plot will
        be saved at the specified dir.
    show_plot: boolean
        Whether to show the plot or not.
    """
    ylabeldist = -10
    electrode_id = 0
    cmap = plt.get_cmap("tab10")
    for k, dev in enumerate(configs['devices']):
        fig, axs = plt.subplots(2, 4)
        # plt.grid(True)
        fig.suptitle('Device ' + dev + ' - Input voltage vs Output current')
        for i in range(2):
            for j in range(4):
                exp_index = j + i * 4
                exp = "IV" + str(exp_index + 1)
                if exp_index < 7:
                    if configs["driver"]['instruments_setup'][dev][
                            "activation_channel_mask"][exp_index] == 1:
                        masked_idx = sum(
                            configs["driver"]['instruments_setup'][dev]
                            ["activation_channel_mask"][:exp_index + 1]) - 1
                        # Modifying x-axis
                        temp = inputs[exp][dev][masked_idx]
                        if not configs['driver']['instruments_setup'][
                                'average_io_point_difference']:
                            wm = WaveformManager({
                                'slope_length':
                                0,
                                'plateau_length':
                                int(configs['driver']['instruments_setup']
                                    ['readout_sampling_frequency'] /
                                    configs['driver']['instruments_setup']
                                    ['activation_sampling_frequency'])
                            })
                            temp = wm.points_to_plateaus(
                                torch.tensor(
                                    inputs[exp][dev]
                                    [masked_idx])).detach().cpu().numpy()
                        axs[i, j].plot(temp,
                                       output[exp][dev],
                                       color=cmap(exp_index))
                        axs[i, j].set_ylabel('output (nA)',
                                             labelpad=ylabeldist)
                        axs[i, j].set_xlabel('input (V)', labelpad=1)
                        axs[i, j].xaxis.grid(True)
                        axs[i, j].yaxis.grid(True)
                    else:
                        # if self.configs["driver"]['instruments_setup'][
                        #         dev]["activation_channel_mask"][
                        #             exp_index] == 1:
                        #     axs[i,
                        #         j].plot(input_waveform[exp_index]
                        #                 [:, electrode_id])

                        #     axs[i, j].set_title(
                        #         devlist[dev]["activation_channels"]
                        #         [exp_index])
                        axs[i, j].plot([])
                        axs[i, j].set_xlabel('Channel Masked')
                    electrode_id += 1
                else:
                    for z, key in enumerate(inputs.keys()):
                        m = 0
                        if configs["driver"]['instruments_setup'][dev][
                                "activation_channel_mask"][z] == 1:
                            masked_idx = sum(
                                configs["driver"]['instruments_setup'][dev]
                                ["activation_channel_mask"][:z + 1]) - 1
                            temp = inputs[key][dev][masked_idx]
                            if not configs['driver']['instruments_setup'][
                                    'average_io_point_difference']:
                                wm = WaveformManager({
                                    'slope_length':
                                    0,
                                    'plateau_length':
                                    int(configs['driver']['instruments_setup']
                                        ['readout_sampling_frequency'] /
                                        configs['driver']['instruments_setup']
                                        ['activation_sampling_frequency'])
                                })
                                temp = wm.points_to_plateaus(
                                    torch.tensor(inputs[key][dev][masked_idx])
                                ).detach().cpu().numpy(
                                )[int(configs['driver']['instruments_setup']
                                      ['readout_sampling_frequency'] /
                                      configs['driver']['instruments_setup']
                                      ['activation_sampling_frequency']):]
                            axs[i, j].plot(temp,
                                           label="IV" + str(z),
                                           color=cmap(z))
                            m += 1
                    #axs[i, j].yaxis.tick_right()
                    #axs[i, j].yaxis.set_label_position("right")
                    axs[i, j].set_ylabel('input (V)')
                    axs[i, j].set_xlabel('points', labelpad=1)
                    axs[i, j].set_title("Input input_signal")
                    axs[i, j].xaxis.grid(True)
                    axs[i, j].yaxis.grid(True)
                    axs[i, j].legend()
    plt.subplots_adjust(hspace=0.3, wspace=0.35)
    if save_plot is not None:
        plt.savefig(save_plot)
    if show_plot:
        plt.show()



import unittest
from bspysmg.utils import plots
import numpy as np
import matplotlib.pyplot as plt
from brainspy.utils.io import load_configs

class Test_Plots(unittest.TestCase):
    def test_error_hist(self):

        target = np.random.uniform(0, 1, size=100)
        predictions = np.random.uniform(0, 1, size=100)
        error = target - predictions
        mse = np.mean(error**2)

        try:
            plots.plot_error_hist(target, predictions, error, mse, ".")
        except:
            self.fail("Failed in creating error hist")

        with self.assertRaises(AttributeError):
            plots.plot_error_hist(list(target), list(predictions), list(error),
                                  mse, ".")

        with self.assertRaises(TypeError):
            plots.plot_error_hist()

        with self.assertRaises(AssertionError):
            plots.plot_error_hist(target, predictions, np.array([0]), mse, ".")

        with self.assertRaises(AssertionError):
            plots.plot_error_hist(target, predictions, error, -1, ".")
        plt.close("all")

    def test_error_output(self):

        target = np.random.uniform(0, 1, size=100)
        error = np.random.uniform(0, 1, size=100)

        try:
            plots.plot_error_vs_output(target, error, ".")
        except:
            self.fail("Failed in creating error vs output hist")

        with self.assertRaises(AttributeError):
            plots.plot_error_vs_output(list(target), list(error), ".")

        with self.assertRaises(TypeError):
            plots.plot_error_vs_output()

        with self.assertRaises(AssertionError):
            plots.plot_error_vs_output(target, np.array([0]), ".")
        plt.close("all")

    def test_output_hist(self):

        output = np.random.uniform(0, 1, size=100)

        try:
            plots.output_hist(output, ".", 5)
            plots.output_hist(output, ".", 10, False)
            plots.output_hist(output, ".")
            plots.output_hist(list(output), ".")
        except:
            self.fail("Fail in creating output hist")

        with self.assertRaises(TypeError):
            plots.output_hist()

        with self.assertRaises(TypeError):
            plots.output_hist(output, ".", 10.1)
        plt.close("all")

    def test_iv_plot(self):

        inputs = np.random.randn(100, 7)
        result = np.random.randn(100, 1)

        try:
            plots.iv_plot(result=result, inputs=inputs, input_electrode=1)
            #plots.iv_plot(inputs, result, ".", True)
            #plots.iv_plot(list(result), 2, ".", True, True)
        except Exception:
            self.fail("Fail in creating IV curve.")

        with self.assertRaises(TypeError):
            plots.iv_plot()

        with self.assertRaises(TypeError):
            plots.iv_plot(result, 5)

        with self.assertRaises(ValueError):
            plots.iv_plot(result, 10.1, ",")
        plt.close("all")

    def test_plot_waves(self):

        inp = np.random.uniform(0, 1, size=100)
        out = np.random.uniform(0, 1, size=100)
        legend = np.array(["Training", "Training"])

        try:
            plots.plot_waves(inp, out, 0, 1, 1, legend, "./")
            plots.plot_waves(inp, out, 0, 1, 1.1, legend, "./")
            plots.plot_waves(list(inp), list(out), 0, 1, 1.1, legend, "./")
        except:
            self.fail("Fail in creating input and output waves")

        with self.assertRaises(TypeError):
            plots.plot_waves()
        plt.close("all")

    def test_multiple_plots(self):
        configs = load_configs(
            'configs/utils/brains_ivcurve_template.yaml')
        inputs = load_configs('tests/data/ins.yaml')
        output = load_configs('tests/data/outs.yaml')
        configs['driver']['instruments_setup']['activation_sampling_frequency'] = 100
        configs['driver']['instruments_setup']['readout_sampling_frequency'] = 100
        configs['driver']['instruments_setup']['average_io_point_difference'] = True
        plots.multi_iv_plot(configs, inputs, output, show_plot=False)
        configs['driver']['instruments_setup']['average_io_point_difference'] = False
        plots.multi_iv_plot(configs, inputs, output, show_plot=False)


if __name__ == '__main__':
    unittest.main()



import unittest
from brainspy.processors.processor import Processor
from brainspy.utils.pytorch import TorchUtils
from bspysmg.utils import consistency
from bspysmg import TEST_MODE
import numpy as np

class Test_Consistency(unittest.TestCase):
    def __init__(self, *args, **kwargs) -> None:
        super(Test_Consistency, self).__init__(*args, **kwargs)
        configs = {}
        configs = {}
        configs['processor_type'] = 'simulation'
        configs["electrode_effects"] = {}
        configs["waveform"] = {}
        configs["waveform"]["plateau_length"] = 1 #10
        configs["waveform"]["slope_length"] = 0 #30
        self.device_configs = configs

        model_data = {}
        model_data["info"] = {}
        model_data["info"]["model_structure"] = {
            "hidden_sizes": [90, 90, 90],
            "D_in": 7,
            "D_out": 1,
            "activation": "relu",
        }
        model_data["info"]['electrode_info'] = {}
        model_data["info"]['electrode_info']['electrode_no'] = 8
        model_data["info"]['electrode_info']['activation_electrodes'] = {}
        model_data["info"]['electrode_info']['activation_electrodes']['electrode_no'] = 7
        model_data["info"]['electrode_info']['activation_electrodes'][
                'voltage_ranges'] = np.array([[-0.55, 0.325], [-0.95, 0.55],
                                            [-1., 0.6], [-1., 0.6], [-1., 0.6],
                                            [-0.95, 0.55], [-0.55, 0.325]])
        model_data["info"]['electrode_info']['output_electrodes'] = {}
        model_data["info"]['electrode_info']['output_electrodes']['electrode_no'] = 1
        model_data["info"]['electrode_info']['output_electrodes']['amplification'] = [28.5]
        model_data["info"]['electrode_info']['output_electrodes']['clipping_value'] = None
        self.model_data = model_data

    @unittest.skipUnless(TEST_MODE == "HARDWARE_CDAQ"
                         or TEST_MODE == "HARDWARE_NIDAQ",
                         "Hardware test is skipped for simulation setup.")
    def test_consistency(self):
        try:
            consistency.consistency_check(
                    'tests/data/',
                    repetitions=1,
                    charge_device=True,
                    show_plots=False)
        except Exception:
            self.fail("Not possible to run consistency check")

    @unittest.skipUnless(TEST_MODE == "HARDWARE_CDAQ",
                         "Hardware test is skipped for simulation setup.")
    def test_consistency_with_model(self):
        model = Processor(self.device_configs, self.model_data['info'])
        model = TorchUtils.format(model)
        try:
            consistency.consistency_check('tests/data',
                    repetitions=1,
                    charge_device=True,
                    show_plots=False,
                    model=model)
        except Exception:
            self.fail("Not possible to run consistency check")



if __name__ == '__main__':
    unittest.main()


from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="brainspy-smg",
    version="1.0.7",
    description=
    "A python package to support research on different nano-scale materials for creating hardware accelerators in the context of deep neural networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['pytorch', 'nano', 'material', 'hardware acceleration', 'deep neural networks', 'surrogate model', 'twin device'],	
    author="Unai Alegre-Ibarra et al.",
    author_email="u.alegre@utwente.nl",
    license="GPL-3.0",
    #python_requires='==3.9',
    install_requires=["brainspy", "scipy~=1.5", "more_itertools==8.6.0"],
    packages=find_packages(exclude=["tests"]),
    url="https://github.com/BraiNEdarwin/brainspy-smg",
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ]
)



import unittest
import torch
from bspysmg.data import dataset
from brainspy.utils.io import load_configs


class Test_Dataset(unittest.TestCase):
    def test_empty_inputs(self):
        with self.assertRaises(TypeError):
            dataset.ModelDataset()

    def test_dataset(self):
        try:
            ds = dataset.ModelDataset("tests/data/postprocessed_data.npz")
            dl = torch.utils.data.DataLoader(ds, batch_size=100)
            next(iter(dl))
            len(dl)
        except Exception:
            self.fail('Dataloader could not be used with the Model Dataset')

    def test_normal_inputs(self):
        try:
            dataset.ModelDataset("tests/data/postprocessed_data.npz", steps=10)
        except Exception:
            self.fail("Failed Execution: Steps - 10, Tag - train")
        try:
            dataset.ModelDataset("tests/data/postprocessed_data.npz", steps=5)
        except Exception:
            self.fail("Failed Execution: Steps - 5, Tag - validation")
        try:
            dataset.ModelDataset("tests/data/postprocessed_data.npz", steps=1)
        except Exception:
            self.fail("Failed Execution: Steps - 1, Tag - test")

    def test_check_file_exists(self):
        with self.assertRaises(FileNotFoundError):
            dataset.ModelDataset("test_fail.npz")

    def test_steps_values(self):
        with self.assertRaises(TypeError):
            dataset.ModelDataset("tests/data/postprocessed_data.npz",
                                 steps=0.1)
        with self.assertRaises(TypeError):
            dataset.ModelDataset("tests/data/postprocessed_data.npz",
                                 steps="10")
        with self.assertRaises(TypeError):
            dataset.ModelDataset("tests/data/postprocessed_data.npz",
                                 steps=[1, 2, 3])

    def test_get_dataloaders_three_paths(self):
        configs = load_configs("configs/training/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz"
        ] * 3
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail(
                "Execution Failed: for three dataset paths (train, val and test)"
            )

    def test_get_dataloaders_two_paths(self):
        configs = load_configs("tests/data/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz"
        ] * 2
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail(
                "Execution Failed: for two dataset paths (train and val)")

    def test_get_dataloaders_one_path(self):
        configs = load_configs("tests/data/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz"
        ]
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail("Execution Failed: for a single dataset path (train)")

    def test_get_dataloaders_one_path_one_split(self):
        configs = load_configs("tests/data/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz"
        ]
        configs['data']['split_percentages'] = [1]
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail("Execution Failed: for a single dataset path (train)")

    def test_get_dataloaders_one_path_two_split(self):
        configs = load_configs("tests/data/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz"
        ]
        configs['data']['split_percentages'] = [0.8, 0.2]
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail("Execution Failed: for a single dataset path (train)")

    def test_get_dataloaders_two_path_one_none(self):
        configs = load_configs("configs/training/smg_configs_template.yaml")
        configs["data"]["dataset_paths"] = [
            "tests/data/postprocessed_data.npz", None
        ]
        try:
            dataset.get_dataloaders(configs)
        except Exception:
            self.fail("Execution Failed: for a single dataset path (train)")

    def test_get_dataloaders_assertions(self):
        configs = load_configs("tests/data/smg_configs_template.yaml")
        with self.assertRaises(AssertionError):
            configs_1 = configs.copy()
            configs_1["data"]["dataset_paths"] = []
            dataset.get_dataloaders(configs_1)

        with self.assertRaises(AssertionError):
            configs_2 = configs.copy()
            configs_2["data"]["dataset_paths"] = "Target"
            dataset.get_dataloaders(configs_2)

        with self.assertRaises(AssertionError):
            configs_3 = configs.copy()
            del configs_3["data"]
            dataset.get_dataloaders(configs_3)

        with self.assertRaises(AssertionError):
            configs_4 = configs.copy()
            del configs_4["data"]["dataset_paths"]
            dataset.get_dataloaders(configs_4)


if __name__ == "__main__":
    unittest.main()